# Copyright (c) 2010, 2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import os
import unittest

import io
import re
from importlib import reload
from . import fixture

from iniparse import RawConfigParser, SafeConfigParser
from unittest.mock import Mock, patch, MagicMock, mock_open
import tempfile
from iniparse import ConfigParser

import rhsm.connection
from subscription_manager import repofile

# repofile must be patched and reloaded to import AptRepofile, otherwise
# the class is not defined in the first place
deb_mock = MagicMock()
deb_mock.Deb822 = dict
with patch.dict("subscription_manager.repofile.sys.modules", {"debian.deb822": deb_mock}):
    reload(repofile)
    from subscription_manager.repofile import AptRepoFile
reload(repofile)

from .stubs import (
    StubProductCertificate,
    StubProduct,
    StubEntitlementCertificate,
    StubContent,
    StubProductDirectory,
    StubConsumerIdentity,
    StubEntitlementDirectory,
    StubDeb822,
)
from subscription_manager.repolib import (
    RepoActionInvoker,
    RepoUpdateActionCommand,
    YumReleaseverSource,
    YumPluginManager,
)
from subscription_manager.repofile import Repo, TidyWriter, YumRepoFile
from subscription_manager import injection as inj
from rhsm.config import RhsmConfigParser
from rhsmlib.services import config

from subscription_manager import repolib
from subscription_manager.entcertlib import CONTENT_ACCESS_CERT_TYPE


# always pretend to run tests with yum in order to have 'ui_repoid_vars' available
repofile.HAS_YUM = True


class ConfigFromString(config.Config):
    def __init__(self, config_string):
        parser = RhsmConfigParserFromString(config_string)
        super(ConfigFromString, self).__init__(parser)


class RhsmConfigParserFromString(RhsmConfigParser):
    def __init__(self, config_string):
        SafeConfigParser.__init__(self)
        self.stringio = io.StringIO(config_string)
        self.readfp(self.stringio)


class TestRepoActionInvoker(fixture.SubManFixture):
    def _stub_content(self, include_content_access=False):
        stub_prod = StubProduct("stub_product", provided_tags="stub-product")

        stub_content = StubContent("a_test_repo", required_tags="stub-product")

        stub_content2 = StubContent("test_repo_2", required_tags="stub-product")

        stub_ent_cert = StubEntitlementCertificate(stub_prod, content=[stub_content])
        stub_prod_cert = StubProductCertificate(stub_prod)

        certs = [stub_ent_cert]
        if include_content_access:
            self.stub_content_access_cert = StubEntitlementCertificate(
                stub_prod, content=[stub_content, stub_content2], entitlement_type=CONTENT_ACCESS_CERT_TYPE
            )
            # content access cert is first and last, so naively wrong implementations will prioritize it.
            certs = [self.stub_content_access_cert, stub_ent_cert, self.stub_content_access_cert]
        stub_ent_dir = StubEntitlementDirectory(certs)
        stub_prod_dir = StubProductDirectory([stub_prod_cert])

        inj.provide(inj.ENT_DIR, stub_ent_dir)
        inj.provide(inj.PROD_DIR, stub_prod_dir)

    def test_is_managed(self):
        self._stub_content()
        repo_action_invoker = RepoActionInvoker()
        repo_label = "a_test_repo"

        im_result = repo_action_invoker.is_managed(repo_label)

        self.assertTrue(im_result)

    def test_get_repo_file(self):
        repo_action_invoker = RepoActionInvoker()

        repo_file = repo_action_invoker.get_repo_file()
        self.assertFalse(repo_file is None)

    def test_get_repos_empty_dirs(self):
        repo_action_invoker = RepoActionInvoker()
        repos = repo_action_invoker.get_repos()
        if repos:
            self.fail("get_repos() should have returned an empty set but did not.")

    def test_get_repos(self):
        self._stub_content(include_content_access=True)
        repo_action_invoker = RepoActionInvoker()
        repos = repo_action_invoker.get_repos()
        self.assertEqual(2, len(repos), "Should produce two repos")
        matching_repos = [repo for repo in repos if repo.id == "a_test_repo"]
        self.assertEqual(1, len(matching_repos), 'Should only produce one repo for "a_test_repo"')
        repo = matching_repos[0]
        certpath = repo.get("sslclientcert")
        self.assertNotEqual(certpath, self.stub_content_access_cert.path)


PROXY_NO_PROTOCOL = """
[server]
proxy_hostname = fake.server.com
proxy_port = 3129
"""

PROXY_EMPTY_PROTOCOL = """
[server]
proxy_hostname = fake.server.com
proxy_port = 3129
proxy_scheme =
"""

PROXY_HTTP_PROTOCOL = """
[server]
proxy_hostname = fake.server.com
proxy_scheme = http
proxy_port = 3129
"""

PROXY_HTTPS_PROTOCOL = """
[server]
proxy_hostname = fake.server.com
proxy_scheme = https
proxy_port = 3129
"""

PROXY_EXTRA_SCHEME = """
[server]
proxy_hostname = fake.server.com
proxy_scheme = https://
proxy_port = 3129
"""


class RepoTests(unittest.TestCase):
    """
    Tests for the repolib Repo class
    """

    def test_valid_label_for_id(self):
        repo_id = "valid-label"
        repo = Repo(repo_id)
        self.assertEqual(repo_id, repo.id)

    def test_valid_unicode_just_ascii_label_for_id(self):
        repo_id = "valid-label"
        repo = Repo(repo_id)
        self.assertEqual(repo_id, repo.id)

    def test_invalid_unicode_label_for_id(self):
        repo_id = "valid-不明-label"
        repo = Repo(repo_id)
        expected = "valid----label"
        self.assertEqual(expected, repo.id)

    def test_invalid_label_with_spaces(self):
        repo_id = "label with spaces"
        repo = Repo(repo_id)
        self.assertEqual("label-with-spaces", repo.id)

    def test_existing_order_is_preserved(self):
        config = (("key 1", "value 1"), ("key b", "value b"), ("key 3", "value 3"))
        repo = Repo("testrepo", config)
        self.assertEqual(config, tuple(repo.items())[:3])

    def test_empty_strings_not_set_in_file(self):
        r = Repo("testrepo", (("proxy", ""),))
        r["proxy"] = ""
        self.assertFalse(("proxy", "") in list(r.items()))

    def test_unknown_property_is_preserved(self):
        existing_repo = Repo("testrepo")
        existing_repo["fake_prop"] = "fake"
        self.assertTrue(("fake_prop", "fake") in list(existing_repo.items()))

    @patch.object(repofile, "conf", ConfigFromString(config_string=PROXY_NO_PROTOCOL))
    def test_http_by_default(self):
        repo = Repo("testrepo")
        r = Repo._set_proxy_info(repo)
        self.assertEqual(r["proxy"], "http://fake.server.com:3129")

    @patch.object(repofile, "conf", ConfigFromString(config_string=PROXY_EMPTY_PROTOCOL))
    def test_http_by_empty(self):
        repo = Repo("testrepo")
        r = Repo._set_proxy_info(repo)
        self.assertEqual(r["proxy"], "http://fake.server.com:3129")

    @patch.object(repofile, "conf", ConfigFromString(config_string=PROXY_HTTP_PROTOCOL))
    def test_http(self):
        repo = Repo("testrepo")
        r = Repo._set_proxy_info(repo)
        self.assertEqual(r["proxy"], "http://fake.server.com:3129")

    @patch.object(repofile, "conf", ConfigFromString(config_string=PROXY_HTTPS_PROTOCOL))
    def test_https(self):
        repo = Repo("testrepo")
        r = Repo._set_proxy_info(repo)
        self.assertEqual(r["proxy"], "https://fake.server.com:3129")

    @patch.object(repofile, "conf", ConfigFromString(config_string=PROXY_EXTRA_SCHEME))
    def test_extra_chars_in_scheme(self):
        repo = Repo("testrepo")
        r = Repo._set_proxy_info(repo)
        self.assertEqual(r["proxy"], "https://fake.server.com:3129")


class RepoActionReportTests(fixture.SubManFixture):
    def test(self):
        report = repolib.RepoActionReport()
        repo = self._repo("a-unicode-content-label", "A unicode repo name")
        report.repo_updates.append(repo)
        report.repo_added.append(repo)
        deleted_section = "einige-repo-name"
        deleted_section_2 = "一些回購名稱"
        report.repo_deleted = [deleted_section, deleted_section_2]

        str(report)

        with fixture.locale_context("de_DE.utf8"):
            str(report)

    def _repo(self, id, name):
        repo = repolib.Repo(repo_id=id)
        repo["name"] = name
        return repo

    def test_empty(self):
        report = repolib.RepoActionReport()
        self.assertEqual(report.updates(), 0)
        "%s" % report
        str(report)

    def test_format(self):
        report = repolib.RepoActionReport()
        repo = self._repo("a-repo-label", "A Repo Name")
        report.repo_updates.append(repo)
        res = str(report)
        # needs tests run in eng locale, coupled to report format since
        # I managed to typo them
        report_label_regexes = [
            "^Repo updates$",
            "^Total repo updates:",
            "^Updated$",
            r"^Added \(new\)$",
            "^Deleted$",
        ]
        for report_label_regex in report_label_regexes:
            if not re.search(report_label_regex, res, re.MULTILINE):
                self.fail("Expected to match the report label regex  %s but did not" % report_label_regex)


class RepoUpdateActionTests(fixture.SubManFixture):
    def setUp(self):
        super(RepoUpdateActionTests, self).setUp()
        stub_prod = StubProduct("fauxprod", provided_tags="TAG1,TAG2")
        stub_prod2 = StubProduct("fauxprovidedprod", provided_tags="TAG4")
        stub_prod_cert = StubProductCertificate(stub_prod, provided_products=[stub_prod2])
        stub_prod2 = StubProduct("fauxprod2", provided_tags="TAG5,TAG6")
        stub_prod2_cert = StubProductCertificate(stub_prod2)
        prod_dir = StubProductDirectory([stub_prod_cert, stub_prod2_cert])
        inj.provide(inj.PROD_DIR, prod_dir)

        stub_content = [
            StubContent("c1", required_tags="", gpg=None),  # no required tags
            StubContent("c2", required_tags="TAG1", gpg=""),
            StubContent("c3", required_tags="TAG1,TAG2,TAG3"),  # should get skipped
            StubContent("c4", required_tags="TAG1,TAG2,TAG4,TAG5,TAG6", gpg="/gpg.key", url="/$some/$path"),
            StubContent("c5", content_type="file", required_tags="", gpg=None),
            StubContent("c6", content_type="file", required_tags="", gpg=None),
        ]
        self.stub_ent_cert = StubEntitlementCertificate(stub_prod, content=stub_content)
        ent_dir = StubEntitlementDirectory([self.stub_ent_cert])
        inj.provide(inj.ENT_DIR, ent_dir)

        repolib.ConsumerIdentity = StubConsumerIdentity

    def _find_content(self, content_list, name):
        """
        Scan list of content for one with name.
        """
        for content in content_list:
            if content["name"] == name:
                return content
        return None

    def test_override_cache_update_skipped_when_overrides_not_supported_on_server(self):
        inj.provide(inj.OVERRIDE_STATUS_CACHE, Mock())
        override_cache_mock = inj.require(inj.OVERRIDE_STATUS_CACHE)

        mock_uep = Mock()
        mock_uep.supports_resource = Mock(return_value=False)
        mock_uep.get_supported_resources = Mock(return_value=[])
        mock_uep.getCertificates = Mock(return_value=[])
        mock_uep.getCertificateSerials = Mock(return_value=[])
        mock_uep.getRelease = Mock(return_value={"releaseVer": "dummyrelease"})

        self.set_consumer_auth_cp(mock_uep)

        RepoUpdateActionCommand()

        # No cache calls should be made if overrides are not supported
        self.assertFalse(override_cache_mock.read_cache.called)
        self.assertFalse(override_cache_mock.load_status.called)

    def test_overrides_trump_ent_cert(self):
        update_action = RepoUpdateActionCommand()
        update_action.overrides = {"x": {"gpgcheck": "blah"}}
        r = Repo("x", [("gpgcheck", "original"), ("gpgkey", "some_key")])
        self.assertEqual("original", r["gpgcheck"])
        update_action._set_override_info(r)
        self.assertEqual("blah", r["gpgcheck"])
        self.assertEqual("some_key", r["gpgkey"])

    def test_overrides_trump_existing(self):
        update_action = RepoUpdateActionCommand()
        update_action.overrides = {"x": {"gpgcheck": "blah"}}
        values = [("gpgcheck", "original"), ("gpgkey", "some_key")]
        old_repo = Repo("x", values)
        new_repo = Repo(old_repo.id, values)
        update_action._set_override_info(new_repo)
        self.assertEqual("original", old_repo["gpgcheck"])
        update_action.update_repo(old_repo, new_repo)
        self.assertEqual("blah", old_repo["gpgcheck"])
        self.assertEqual("some_key", old_repo["gpgkey"])

    @patch("subscription_manager.repolib.get_repo_file_classes")
    def test_update_when_new_repo(self, mock_get_repo_file_classes):
        mock_file = MagicMock()
        mock_file.CONTENT_TYPES = [None]
        mock_file.fix_content = lambda x: x
        mock_file.section.return_value = None
        mock_class = MagicMock(return_value=mock_file)
        mock_get_repo_file_classes.return_value = [(mock_class, mock_class)]

        def stub_content():
            return [Repo("x", [("gpgcheck", "original"), ("gpgkey", "some_key")])]

        update_action = RepoUpdateActionCommand()
        update_action.get_unique_content = stub_content
        update_report = update_action.perform()
        written_repo = mock_file.add.call_args[0][0]
        self.assertEqual("original", written_repo["gpgcheck"])
        self.assertEqual("some_key", written_repo["gpgkey"])
        self.assertEqual(1, update_report.updates())

    @patch("subscription_manager.repolib.get_repo_file_classes")
    def test_update_when_repo_not_modified_on_mutable(self, mock_get_repo_file_classes):
        self._inject_mock_invalid_consumer()
        modified_repo = Repo("x", [("gpgcheck", "original"), ("gpgkey", "some_key")])
        server_repo = Repo("x", [("gpgcheck", "original")])
        mock_file = MagicMock()
        mock_file.CONTENT_TYPES = [None]
        mock_file.fix_content = lambda x: x
        mock_file.section.side_effect = [modified_repo, server_repo]
        mock_class = MagicMock(return_value=mock_file)
        mock_get_repo_file_classes.return_value = [(mock_class, mock_class)]

        def stub_content():
            return [Repo("x", [("gpgcheck", "new"), ("gpgkey", "new_key"), ("name", "test")])]

        update_action = RepoUpdateActionCommand()
        update_action.get_unique_content = stub_content
        current = update_action.perform()
        # confirming that the assessed value does change when repo file
        # is the same as the server value file.
        self.assertEqual("new", current.repo_updates[0]["gpgcheck"])

        # this is the ending server value file.
        written_repo = mock_file.update.call_args[0][0]
        self.assertEqual("new", written_repo["gpgcheck"])
        self.assertEqual(None, written_repo["gpgkey"])

    @patch("subscription_manager.repolib.get_repo_file_classes")
    def test_update_when_repo_modified_on_mutable(self, mock_get_repo_file_classes):
        self._inject_mock_invalid_consumer()
        modified_repo = Repo("x", [("gpgcheck", "unoriginal"), ("gpgkey", "some_key")])
        server_repo = Repo("x", [("gpgcheck", "original")])
        mock_file = MagicMock()
        mock_file.CONTENT_TYPES = [None]
        mock_file.fix_content = lambda x: x
        mock_file.section.side_effect = [modified_repo, server_repo]
        mock_class = MagicMock(return_value=mock_file)
        mock_get_repo_file_classes.return_value = [(mock_class, mock_class)]

        def stub_content():
            return [Repo("x", [("gpgcheck", "new"), ("gpgkey", "new_key"), ("name", "test")])]

        update_action = RepoUpdateActionCommand()
        update_action.get_unique_content = stub_content
        current = update_action.perform()
        # confirming that the assessed value does not change when repo file
        # is different from the server value file.
        self.assertEqual("unoriginal", current.repo_updates[0]["gpgcheck"])

        # this is the ending server value file
        written_repo = mock_file.update.call_args[0][0]
        self.assertEqual("new", written_repo["gpgcheck"])
        self.assertEqual(None, written_repo["gpgkey"])

    def test_no_gpg_key(self):

        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        c1 = self._find_content(content, "c1")
        self.assertEqual("", c1["gpgkey"])
        self.assertEqual("0", c1["gpgcheck"])

        c2 = self._find_content(content, "c2")
        self.assertEqual("", c2["gpgkey"])
        self.assertEqual("0", c2["gpgcheck"])

    def test_gpg_key(self):

        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        c4 = self._find_content(content, "c4")
        self.assertEqual("http://example.com/gpg.key", c4["gpgkey"])
        self.assertEqual("1", c4["gpgcheck"])

    def test_ui_repoid_vars(self):
        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        c4 = self._find_content(content, "c4")
        self.assertEqual("some path", c4["ui_repoid_vars"])
        c2 = self._find_content(content, "c2")
        self.assertEqual(None, c2["ui_repoid_vars"])

    def test_tags_found(self):
        update_action = RepoUpdateActionCommand()
        content = update_action.get_unique_content()
        self.assertEqual(3, len(content))

    def test_only_allow_content_of_type_yum(self):
        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        self.assertTrue(self._find_content(content, "c1") is not None)
        self.assertTrue(self._find_content(content, "c5") is None)
        self.assertTrue(self._find_content(content, "c6") is None)

    def test_mutable_property(self):
        update_action = RepoUpdateActionCommand()
        self._inject_mock_invalid_consumer()
        existing_repo = Repo("testrepo")
        existing_repo["metadata_expire"] = 1000
        incoming_repo = {"metadata_expire": 2000}
        update_action.update_repo(existing_repo, incoming_repo)
        self.assertEqual(1000, existing_repo["metadata_expire"])

    def test_mutable_property_is_server(self):
        update_action = RepoUpdateActionCommand()
        self._inject_mock_invalid_consumer()
        existing_repo = Repo("testrepo")
        server_val_repo = Repo("servertestrepo")
        existing_repo["metadata_expire"] = 1000
        server_val_repo["metadata_expire"] = 1000
        incoming_repo = {"metadata_expire": 2000}
        update_action.update_repo(existing_repo, incoming_repo, server_val_repo)
        self.assertEqual(2000, existing_repo["metadata_expire"])

    def test_gpgcheck_is_mutable(self):
        update_action = RepoUpdateActionCommand()
        self._inject_mock_invalid_consumer()
        existing_repo = Repo("testrepo")
        existing_repo["gpgcheck"] = "0"
        incoming_repo = {"gpgcheck": "1"}
        update_action.update_repo(existing_repo, incoming_repo)
        self.assertEqual("0", existing_repo["gpgcheck"])

    def test_mutable_property_in_repo_but_not_in_cert(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["metadata_expire"] = 1000
        incoming_repo = {}
        update_action.update_repo(existing_repo, incoming_repo)
        self.assertEqual(1000, existing_repo["metadata_expire"])

    def test_immutable_property(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["name"] = "meow"
        incoming_repo = {"name": "woof"}
        update_action.update_repo(existing_repo, incoming_repo)
        self.assertEqual("woof", existing_repo["name"])

    # If the user removed a mutable property completely, or the property
    # is new in a new version of the entitlement certificate, the new value
    # should get written out.
    def test_unset_mutable_property(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        incoming_repo = {"metadata_expire": 2000}
        update_action.update_repo(existing_repo, incoming_repo)
        self.assertEqual(2000, existing_repo["metadata_expire"])

    def test_unset_immutable_property(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        incoming_repo = {"name": "woof"}
        update_action.update_repo(existing_repo, incoming_repo)
        self.assertEqual("woof", existing_repo["name"])

    # Test repo on disk has an immutable property set which has since been
    # unset in the new repo definition. This property should be removed.
    def test_set_immutable_property_now_empty(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["proxy_username"] = "blah"
        incoming_repo = {}
        update_action.update_repo(existing_repo, incoming_repo)
        self.assertFalse("proxy_username" in list(existing_repo.keys()))

    def test_set_mutable_property_now_empty_value(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["metadata_expire"] = "blah"
        incoming_repo = {"metadata_expire": ""}
        update_action.update_repo(existing_repo, incoming_repo)
        # re comments in repolib
        # Mutable properties should be added if not currently defined,
        # otherwise left alone.
        self.assertTrue("metadata_expire" in list(existing_repo.keys()))

    def test_set_custom_property_removed(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["exclude"] = "1"
        # 'exclude' is not in the updated repo listing.
        # this is how an override delete is passed into
        # the update method
        incoming_repo = Repo("testrepo")
        update_action.update_repo(existing_repo, incoming_repo, existing_repo)
        self.assertFalse("exclude" in list(existing_repo.keys()))

    def test_set_immutable_property_now_empty_value(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["proxy_username"] = "blah"
        incoming_repo = {"proxy_username": ""}
        update_action.update_repo(existing_repo, incoming_repo)
        # Immutable properties should be always be added/updated,
        # and removed if undefined in the new repo definition.
        self.assertFalse("proxy_username" in list(existing_repo.keys()))

    def test_set_mutable_property_now_none(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["metadata_expire"] = "blah"
        incoming_repo = {"metadata_expire": None}
        update_action.update_repo(existing_repo, incoming_repo)
        # re comments in repolib
        # Mutable properties should be added if not currently defined,
        # otherwise left alone.
        self.assertTrue("metadata_expire" in list(existing_repo.keys()))

    def test_set_mutable_property_now_not_in_cert(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["metadata_expire"] = "blah"
        incoming_repo = {}
        update_action.update_repo(existing_repo, incoming_repo)
        # re comments in repolib
        # Mutable properties should be added if not currently defined,
        # otherwise left alone.
        self.assertTrue("metadata_expire" in list(existing_repo.keys()))

    def test_set_immutable_property_now_none(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["proxy_username"] = "blah"
        incoming_repo = {"proxy_username": None}
        update_action.update_repo(existing_repo, incoming_repo)
        # Immutable properties should be always be added/updated,
        # and removed if undefined in the new repo definition.
        self.assertFalse("proxy_username" in list(existing_repo.keys()))

    def test_set_immutable_property_now_not_in_cert(self):
        self._inject_mock_invalid_consumer()
        update_action = RepoUpdateActionCommand()
        existing_repo = Repo("testrepo")
        existing_repo["proxy_username"] = "blah"
        incoming_repo = {}
        update_action.update_repo(existing_repo, incoming_repo)
        # Immutable properties should be always be added/updated,
        # and removed if undefined in the new repo definition.
        self.assertFalse("proxy_username" in list(existing_repo.keys()))

    def test_overrides_removed_revert_to_default(self):
        update_action = RepoUpdateActionCommand()
        update_action.written_overrides.overrides = {"x": {"gpgcheck": "blah"}}
        update_action.overrides = {}
        old_repo = Repo("x", [("gpgcheck", "blah"), ("gpgkey", "some_key")])
        new_repo = Repo(old_repo.id, [("gpgcheck", "original"), ("gpgkey", "some_key")])
        update_action._set_override_info(new_repo)
        # The value from the current repo file (with the old override) should exist pre-update
        self.assertEqual("blah", old_repo["gpgcheck"])
        update_action.update_repo(old_repo, new_repo)
        # Because the override has been removed, the value is reset to the default
        self.assertEqual("original", old_repo["gpgcheck"])
        self.assertEqual("some_key", old_repo["gpgkey"])

    def test_overrides_removed_and_edited(self):
        update_action = RepoUpdateActionCommand()
        update_action.written_overrides.overrides = {"x": {"gpgcheck": "override_value"}}
        update_action.overrides = {}
        old_repo = Repo("x", [("gpgcheck", "hand_edit"), ("gpgkey", "some_key")])
        new_repo = Repo(old_repo.id, [("gpgcheck", "original"), ("gpgkey", "some_key")])
        update_action._set_override_info(new_repo)
        # The value from the current repo file (with the old hand edit) should exist pre-update
        self.assertEqual("hand_edit", old_repo["gpgcheck"])
        update_action.update_repo(old_repo, new_repo)
        # Because the current value doesn't match the override, we don't modify it
        self.assertEqual("hand_edit", old_repo["gpgcheck"])
        self.assertEqual("some_key", old_repo["gpgkey"])

    def test_non_default_overrides_added_to_existing(self):
        """
        Test that overrides for values that aren't found in Repo.PROPERTIES are written
        to existing repos
        """
        update_action = RepoUpdateActionCommand()
        update_action.written_overrides.overrides = {}
        update_action.overrides = {"x": {"somekey": "someval"}}
        old_repo = Repo("x", [])
        new_repo = Repo(old_repo.id, [])
        update_action._set_override_info(new_repo)
        update_action.update_repo(old_repo, new_repo)
        self.assertEqual("someval", old_repo["somekey"])

    def test_non_default_override_removed_deleted(self):
        """
        Test that overrides for values that aren't found in Repo.PROPERTIES are
        removed from redhat.repo once the override is removed
        """
        update_action = RepoUpdateActionCommand()
        update_action.written_overrides.overrides = {"x": {"somekey": "someval"}}
        update_action.overrides = {}
        old_repo = Repo("x", [("somekey", "someval")])
        new_repo = Repo(old_repo.id, [])
        update_action._set_override_info(new_repo)
        update_action.update_repo(old_repo, new_repo)
        self.assertFalse("somekey" in old_repo)

    @patch.object(RepoUpdateActionCommand, "get_consumer_auth_cp")
    def test_no_ssl_verify_status(self, mock_get_consumer_auth_cp):
        mock_cp = Mock()
        mock_cp.has_capability = Mock(return_value=False)
        mock_get_consumer_auth_cp.return_value = mock_cp
        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        c1 = self._find_content(content, "c1")
        self.assertIsNone(c1["sslverifystatus"])

    @patch.object(RepoUpdateActionCommand, "get_consumer_auth_cp")
    def test_with_ssl_verify_status(self, mock_get_consumer_auth_cp):
        mock_cp = Mock()
        mock_cp.has_capability = Mock(return_value=True)
        mock_get_consumer_auth_cp.return_value = mock_cp
        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        c1 = self._find_content(content, "c1")
        self.assertEqual("1", c1["sslverifystatus"])

    @patch.object(RepoUpdateActionCommand, "get_consumer_auth_cp")
    def test_with_ssl_verify_ConnectionException(self, mock_get_consumer_auth_cp):
        """Test that network exception is handled and does not use SSL verification."""
        mock_cp = Mock()
        mock_cp.has_capability = Mock(side_effect=rhsm.connection.ConnectionException)
        mock_get_consumer_auth_cp.return_value = mock_cp

        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        c1 = self._find_content(content, "c1")
        self.assertIsNone(c1["sslverifystatus"])

    @patch.object(RepoUpdateActionCommand, "get_consumer_auth_cp")
    def test_with_ssl_verify_ProxyException(self, mock_get_consumer_auth_cp):
        """Test that proxy exception is handled and does not use SSL verification."""
        mock_cp = Mock()
        mock_cp.has_capability = Mock(side_effect=rhsm.connection.ProxyException(hostname="host", port=1234))
        mock_get_consumer_auth_cp.return_value = mock_cp

        update_action = RepoUpdateActionCommand()
        content = update_action.get_all_content(baseurl="http://example.com", ca_cert=None)
        c1 = self._find_content(content, "c1")
        self.assertIsNone(c1["sslverifystatus"])


class TidyWriterTests(unittest.TestCase):
    def test_just_newlines_compressed_to_one(self):
        output = io.StringIO()
        tidy_writer = TidyWriter(output)

        tidy_writer.write("\n\n\n\n")
        tidy_writer.close()

        self.assertEqual("\n", output.getvalue())

    def test_newline_added_to_eof(self):
        output = io.StringIO()
        tidy_writer = TidyWriter(output)

        tidy_writer.write("a line\n")
        tidy_writer.write("another line")
        tidy_writer.close()

        self.assertEqual("a line\nanother line\n", output.getvalue())

    def test_newline_preserved_on_eof(self):
        output = io.StringIO()
        tidy_writer = TidyWriter(output)

        tidy_writer.write("a line\n")
        tidy_writer.write("another line\n")
        tidy_writer.close()

        self.assertEqual("a line\nanother line\n", output.getvalue())

    def test_compression_preserves_a_single_blank_line(self):
        output = io.StringIO()
        tidy_writer = TidyWriter(output)

        tidy_writer.write("test stuff\n\ntest\n")
        tidy_writer.close()

        self.assertEqual("test stuff\n\ntest\n", output.getvalue())

    def test_newlines_compressed_in_single_write(self):
        output = io.StringIO()
        tidy_writer = TidyWriter(output)

        tidy_writer.write("test stuff\n\n\ntest\n")
        tidy_writer.close()

        self.assertEqual("test stuff\n\ntest\n", output.getvalue())

    def test_newlines_compressed_across_writes(self):
        output = io.StringIO()
        tidy_writer = TidyWriter(output)

        tidy_writer.write("test stuff\n\n")
        tidy_writer.write("\ntest\n")
        tidy_writer.close()

        self.assertEqual("test stuff\n\ntest\n", output.getvalue())

        # now try the other split
        output = io.StringIO()
        tidy_writer = TidyWriter(output)

        tidy_writer.write("test stuff\n")
        tidy_writer.write("\n\ntest\n")
        tidy_writer.close()

        self.assertEqual("test stuff\n\ntest\n", output.getvalue())


class YumReleaseverSourceTest(fixture.SubManFixture):
    def test_init(self):
        # inj.provide(inj.RELEASE_STATUS_CACHE, Mock())
        # override_cache_mock = inj.require(inj.OVERRIDE_STATUS_CACHE)

        release_source = YumReleaseverSource()
        self.assertEqual(release_source._expansion, None)
        self.assertEqual(release_source.marker, "$releasever")
        self.assertEqual(release_source.marker, release_source.default)

    def test_default(self):
        release_source = YumReleaseverSource()

        exp = release_source.get_expansion()
        self.assertEqual(exp, "$releasever")

    def test_mem_cache_works(self):
        inj.provide(inj.RELEASE_STATUS_CACHE, Mock())
        release_mock = inj.require(inj.RELEASE_STATUS_CACHE)

        release = "MockServer"
        mock_release = {"releaseVer": release}
        release_mock.read_status = Mock(return_value=mock_release)
        release_source = YumReleaseverSource()

        exp = release_source.get_expansion()
        self.assertEqual(exp, release)
        self.assertEqual(release_source._expansion, release)

        exp = release_source.get_expansion()
        self.assertEqual(exp, release)

    def test_mem_cache_pre_cached(self):
        inj.provide(inj.RELEASE_STATUS_CACHE, Mock())
        release_mock = inj.require(inj.RELEASE_STATUS_CACHE)

        release = "MockServer"
        mock_release = {"releaseVer": release}
        release_mock.read_status = Mock(return_value=mock_release)
        release_source = YumReleaseverSource()

        cached_release = "CachedMockServer"
        release_source._expansion = cached_release
        exp = release_source.get_expansion()
        self.assertEqual(exp, cached_release)
        self.assertEqual(release_source._expansion, cached_release)

    def test_read_status_not_set(self):
        inj.provide(inj.RELEASE_STATUS_CACHE, Mock())
        release_mock = inj.require(inj.RELEASE_STATUS_CACHE)

        release = ""
        mock_release = {"releaseVer": release}
        release_mock.read_status = Mock(return_value=mock_release)
        release_source = YumReleaseverSource()

        exp = release_source.get_expansion()

        # we were unset, should return the default
        self.assertEqual(exp, YumReleaseverSource.default)
        # and cache it
        self.assertEqual(release_source._expansion, YumReleaseverSource.default)


class YumReleaseverSourceIsNotEmptyTest(fixture.SubManFixture):
    def test_empty_string(self):
        self.assertFalse(YumReleaseverSource.is_not_empty(""))

    def test_none(self):
        self.assertFalse(YumReleaseverSource.is_not_empty(None))

    def test_empty_list(self):
        self.assertFalse(YumReleaseverSource.is_not_empty([]))

    def test_number(self):
        self.assertTrue(YumReleaseverSource.is_not_empty("7"))

    def test_string(self):
        self.assertTrue(YumReleaseverSource.is_not_empty("Super"))


class YumReleaseverSourceIsSetTest(fixture.SubManFixture):
    def test_none(self):
        self.assertFalse(YumReleaseverSource.is_set(None))

    def test_key_error(self):
        self.assertFalse(YumReleaseverSource.is_set({"not_release_ver": "blippy"}))

    def test_not_a_dict(self):
        self.assertFalse(YumReleaseverSource.is_set(["some string"]))

    def test_releasever_zero(self):
        self.assertFalse(YumReleaseverSource.is_set({"releaseVer": 0}))

    def test_releasever_none(self):
        self.assertFalse(YumReleaseverSource.is_set({"releaseVer": None}))

    def test_releasever_7server(self):
        self.assertTrue(YumReleaseverSource.is_set({"releaseVer": "7Server"}))

    def test_releasever_the_string_none(self):
        # you laugh, but someone thinks that would be an awesome product name.
        self.assertTrue(YumReleaseverSource.is_set({"releaseVer": "None"}))


class AptRepoFileTest(unittest.TestCase):
    def _helper_stub_repo(self, *args, **kwargs):
        with patch("subscription_manager.repofile.HAS_DEB822", True):
            repo = Repo(*args, **kwargs)
        return repo

    def _helper_stub_repofile(self, *args, **kwargs):
        my_mock = MagicMock()
        with patch("subscription_manager.repofile.RepoFileBase.create", my_mock):
            repofile = AptRepoFile(*args, **kwargs)
        return repofile

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_includes_yum_parameters(self, mock_file):
        # Setup mock
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        ar = self._helper_stub_repofile()
        yr = YumRepoFile()
        mock_repo = self._helper_stub_repo(
            "mock", existing_values=[("baseurl", "https://example.org.org/"), ("sslclientcert", "mypem.pem")]
        )
        # Modify data using different backends (Apt,Yum)
        ar_fixed_content = ar.fix_content(mock_repo)
        yr_fixed_content = yr.fix_content(mock_repo)
        # Yum parameters must be included in Apt as well (garuantee compatibility)
        for key in yr_fixed_content:
            self.assertIn(key, ar_fixed_content)

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_globals(self, mock_file):
        # Setup mock
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        # Setup expected results
        exp_path = "etc/apt/sources.list.d"
        exp_name = "rhsm.sources"
        exp_content_types = ["deb"]
        exp_repofile_header = """#
# Certificate-Based Repositories
# Managed by (rhsm) subscription-manager
#
# *** This file is auto-generated.  Changes made here will be over-written. ***
# *** Use "subscription-manager repo-override --help" if you wish to make changes. ***
#
# If this file is empty and this system is subscribed consider
# a "apt-get update" to refresh available repos
#
# *** DO NOT EDIT THIS FILE ***
#
"""
        ar = self._helper_stub_repofile()
        # Test actual data against expected results
        self.assertEqual(ar.PATH, exp_path)
        self.assertEqual(ar.NAME, exp_name)
        self.assertEqual(ar.CONTENT_TYPES, exp_content_types)
        self.assertEqual(ar.REPOFILE_HEADER, exp_repofile_header)

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_add_repo(self, mock_file):
        # Setup mock
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        ar = self._helper_stub_repofile()
        mock_repo = self._helper_stub_repo("mock")
        mock_dict = dict([(str(k), str(v)) for (k, v) in mock_repo.items()])
        mock_deb822 = StubDeb822(mock_dict)
        # Modify data
        ar.add(mock_repo)
        # Test modification
        act_repo = ar.repos822[0]
        for key in mock_deb822:
            self.assertIn(key, act_repo)
            self.assertEqual(act_repo[key], mock_repo[key])

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_delete_repo(self, mock_file):
        # Setup mock
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        ar = self._helper_stub_repofile()
        mock_repo_a = self._helper_stub_repo("a")
        mock_repo_b = self._helper_stub_repo("b")
        mock_repo_c = self._helper_stub_repo("c")
        ar.add(mock_repo_a)
        ar.add(mock_repo_b)
        ar.add(mock_repo_c)
        # Modify data
        ar.delete(mock_repo_b.id)
        # Test modification
        self.assertEqual(len(ar.repos822), 2)
        self.assertEqual(ar.repos822[0]["id"], "a")
        self.assertEqual(ar.repos822[1]["id"], "c")

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_update_repo(self, mock_file):
        # Setup mock
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        ar = self._helper_stub_repofile()
        mock_repo_a = self._helper_stub_repo("a", existing_values=[("baseurl", "url")])
        ar.add(mock_repo_a)
        # Modify data
        mock_repo_a["baseurl"] = "new url"
        ar.update(mock_repo_a)
        # Test modification
        self.assertEqual(ar.repos822[0]["id"], "a")
        self.assertEqual(ar.repos822[0]["baseurl"], "new url")

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_fix_content_default(self, mock_file):
        # Setup mock and expected values
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        exp_params = {
            "arches": "none",
            "Types": "deb",
            "URIs": "https://example.site.org/",
            "Suites": "default",
            "Components": "all",
            "Trusted": "yes",
            "sslclientcert": "mypem.pem",
        }
        ar = self._helper_stub_repofile()
        repo_mock = self._helper_stub_repo(
            "mock",
            existing_values=[("baseurl", exp_params["URIs"]), ("sslclientcert", exp_params["sslclientcert"])],
        )
        # Modify data
        act_content = ar.fix_content(repo_mock)
        # Test modification by comparing with expected values
        for key in exp_params:
            self.assertIn(key, act_content)
            self.assertEqual(act_content[key], exp_params[key])

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_fix_content_arches(self, mock_file):
        # Setup mock and expected values
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        exp_params = {"arches": "none", "URIs": "https://example.site.org/", "sslclientcert": "mypem.pem"}
        ar = self._helper_stub_repofile()
        repo_mock = self._helper_stub_repo(
            "mock",
            existing_values=[
                ("baseurl", exp_params["URIs"]),
                ("sslclientcert", exp_params["sslclientcert"]),
                ("arches", ["ALL"]),
            ],
        )
        # Modify data
        act_content = ar.fix_content(repo_mock)
        # Test modification by comparing with expected values
        for key in exp_params:
            self.assertIn(key, act_content)
            self.assertEqual(act_content[key], exp_params[key])

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_fix_content_arches_multi(self, mock_file):
        # Setup mock and expected values
        assert open("/etc/apt/sources.list.d/rhsm.sources").read() == "data"
        mock_file.assert_called_with("/etc/apt/sources.list.d/rhsm.sources")
        exp_params = {
            "arches": "amd64 i386",
            "Architectures": "amd64 i386",
            "URIs": "https://example.site.org/",
            "sslclientcert": "mypem.pem",
        }
        ar = self._helper_stub_repofile()
        repo_mock = self._helper_stub_repo(
            "mock",
            existing_values=[
                ("baseurl", exp_params["URIs"]),
                ("sslclientcert", exp_params["sslclientcert"]),
                ("arches", ["amd64", "i386"]),
            ],
        )
        # Modify data
        act_content = ar.fix_content(repo_mock)
        # Test modification by comparing with expected values
        for key in exp_params:
            self.assertIn(key, act_content)
            self.assertEqual(act_content[key], exp_params[key])


class YumRepoFileTest(unittest.TestCase):
    @patch("subscription_manager.repofile.YumRepoFile.create")
    @patch("subscription_manager.repofile.TidyWriter")
    def test_configparsers_equal(self, tidy_writer, stub_create):
        rf = YumRepoFile()
        other = RawConfigParser()
        for parser in [rf, other]:
            parser.add_section("test")
            parser.set("test", "key", "val")
        self.assertTrue(rf._configparsers_equal(other))

    @patch("subscription_manager.repofile.YumRepoFile.create")
    @patch("subscription_manager.repofile.TidyWriter")
    def test_configparsers_diff_sections(self, tidy_writer, stub_create):
        rf = YumRepoFile()
        rf.add_section("new_section")
        other = RawConfigParser()
        for parser in [rf, other]:
            parser.add_section("test")
            parser.set("test", "key", "val")
        self.assertFalse(rf._configparsers_equal(other))

    @patch("subscription_manager.repofile.YumRepoFile.create")
    @patch("subscription_manager.repofile.TidyWriter")
    def test_configparsers_diff_item_val(self, tidy_writer, stub_create):
        rf = YumRepoFile()
        other = RawConfigParser()
        for parser in [rf, other]:
            parser.add_section("test")
            parser.set("test", "key", "val")
        rf.set("test", "key", "val2")
        self.assertFalse(rf._configparsers_equal(other))

    @patch("subscription_manager.repofile.YumRepoFile.create")
    @patch("subscription_manager.repofile.TidyWriter")
    def test_configparsers_diff_items(self, tidy_writer, stub_create):
        rf = YumRepoFile()
        other = RawConfigParser()
        for parser in [rf, other]:
            parser.add_section("test")
            parser.set("test", "key", "val")
        rf.set("test", "somekey", "val")
        self.assertFalse(rf._configparsers_equal(other))

    @patch("subscription_manager.repofile.YumRepoFile.create")
    @patch("subscription_manager.repofile.TidyWriter")
    def test_configparsers_equal_int(self, tidy_writer, stub_create):
        rf = YumRepoFile()
        other = RawConfigParser()
        for parser in [rf, other]:
            parser.add_section("test")
            parser.set("test", "key", "val")
        rf.set("test", "k", 1)
        other.set("test", "k", "1")
        self.assertTrue(rf._configparsers_equal(other))


# config file is root only, so just fill in a stringbuffer
unset_manage_repos_cfg_buf = """
[server]
hostname = server.example.conf
prefix = /candlepin
[rhsm]
manage_repos =

[rhsmcertd]
certCheckInterval = 240
"""

unset_config = """[server]
hostname = server.example.conf
"""

manage_repos_zero_config = """[rhsm]
manage_repos = 0
"""

manage_repos_bool_config = """[rhsm]
manage_repos = false
"""

manage_repos_not_an_int = """[rhsm]
manage_repos = thisisanint
"""

manage_repos_int_37 = """[rhsm]
manage_repos = 37
"""


class TestManageReposEnabled(fixture.SubManFixture):
    @patch.object(repofile, "conf", ConfigFromString(config_string=unset_config))
    def test(self):
        # default stub config, no manage_repo defined, uses default
        manage_repos_enabled = repofile.manage_repos_enabled()
        self.assertEqual(manage_repos_enabled, True)

    @patch.object(repofile, "conf", ConfigFromString(config_string=unset_manage_repos_cfg_buf))
    def test_empty_manage_repos(self):
        manage_repos_enabled = repofile.manage_repos_enabled()
        self.assertEqual(manage_repos_enabled, True)

    @patch.object(repofile, "conf", ConfigFromString(config_string=manage_repos_zero_config))
    def test_empty_manage_repos_zero(self):
        manage_repos_enabled = repofile.manage_repos_enabled()
        self.assertEqual(manage_repos_enabled, False)

    @patch.object(repofile, "config", ConfigFromString(config_string=manage_repos_bool_config))
    def test_empty_manage_repos_bool(self):
        manage_repos_enabled = repofile.manage_repos_enabled()
        # Should fail, and return default of 1
        self.assertEqual(manage_repos_enabled, True)

    @patch.object(repofile, "config", ConfigFromString(config_string=manage_repos_not_an_int))
    def test_empty_manage_repos_not_an_int(self):
        manage_repos_enabled = repofile.manage_repos_enabled()
        # Should fail, and return default of 1
        self.assertEqual(manage_repos_enabled, True)

    @patch.object(repofile, "conf", ConfigFromString(config_string=manage_repos_int_37))
    def test_empty_manage_repos_int_37(self):
        manage_repos_enabled = repofile.manage_repos_enabled()
        # Should fail, and return default of 1
        self.assertEqual(manage_repos_enabled, True)


AUTO_ENABLE_PKG_PLUGINS_ENABLED = """
[rhsm]
auto_enable_yum_plugins = 1
"""

AUTO_ENABLE_PKG_PLUGINS_DISABLED = """
[rhsm]
auto_enable_yum_plugins = 0
"""

PKG_PLUGIN_CONF_FILE_ENABLED_INT = """
[main]
enabled = 1
"""

PKG_PLUGIN_CONF_FILE_ENABLED_BOOL = """
[main]
enabled = true
"""

PKG_PLUGIN_CONF_FILE_DISABLED_INT = """
[main]
enabled = 0
"""

PKG_PLUGIN_CONF_FILE_DISABLED_BOOL = """
[main]
enabled = false
"""

PKG_PLUGIN_CONF_FILE_WRONG_VALUE = """
[main]
enabled = 4
"""

PKG_PLUGIN_CONF_FILE_INVALID_VALUE = """
[main]
enabled = 1.0
"""

PKG_PLUGIN_CONF_MISSING_MAIN_SECTION = """
[test]
option = value
"""

PKG_PLUGIN_CONF_MISSING_ENABLED_OPTION = """
[main]
option = value
"""

PKG_PLUGIN_CONF_FILE_CORRUPTED = """
[main
enable =
"""


class TestYumPluginManager(unittest.TestCase):
    """
    This class is intended for testing YumPluginManager
    """

    ORIGINAL_DNF_PLUGIN_DIR = YumPluginManager.DNF_PLUGIN_DIR
    ORIGINAL_PLUGINS = YumPluginManager.PLUGINS

    def tearDown(self) -> None:
        YumPluginManager.DNF_PLUGIN_DIR = self.ORIGINAL_DNF_PLUGIN_DIR
        YumPluginManager.PLUGINS = self.ORIGINAL_PLUGINS

        self.tempdir = None
        self.file_01 = None
        self.file_02 = None

    def init_dnf_plugin_conf_files(self, conf_string: str):
        """
        Mock configuration files of plugins
        """
        self.tempdir = tempfile.TemporaryDirectory()
        self.file_01 = tempfile.mkstemp(suffix=".conf", dir=self.tempdir.name)
        self.file_02 = tempfile.mkstemp(suffix=".conf", dir=self.tempdir.name)
        os.write(self.file_01[0], conf_string.encode("utf-8"))
        os.write(self.file_02[0], conf_string.encode("utf-8"))
        os.close(self.file_01[0])
        os.close(self.file_02[0])

        YumPluginManager.DNF_PLUGIN_DIR = self.tempdir.name
        YumPluginManager.PLUGINS = [
            self.file_01[1].replace(self.tempdir.name + "/", "").replace(".conf", ""),
            self.file_02[1].replace(self.tempdir.name + "/", "").replace(".conf", ""),
        ]

    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "debian", []),
    )
    def test_enable_pkg_plugins_deb(self, mock_get_distribution):
        """
        Test not enabling of yum plugins on deb based systems (Test 1 of 2)
        """
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 0)

    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "", ["debian"]),
    )
    def test_enable_pkg_plugins_deb_like(self, mock_get_distribution):
        """
        Test not enabling of yum plugins on deb based systems (Test 2 of 2)
        """
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 0)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enable_pkg_plugins_yum(self, mock_get_distribution):
        """
        Test enabling of yum plugins on yum based systems
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_FILE_DISABLED_INT)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 2)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_disabled_dnf_plugin(self, mock_get_distribution):
        """
        Test automatic enabling of configuration files of disabled plugins
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_FILE_DISABLED_INT)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 2)
        for plugin_conf_file_name in YumPluginManager.PLUGINS:
            dnf_plugin_config = ConfigParser()
            result = dnf_plugin_config.read(
                YumPluginManager.DNF_PLUGIN_DIR + "/" + plugin_conf_file_name + ".conf"
            )
            self.assertGreater(len(result), 0)
            is_plugin_enabled = dnf_plugin_config.getint("main", "enabled")
            self.assertEqual(is_plugin_enabled, 1)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_enabled_dnf_plugin_int(self, mock_get_distribution):
        """
        Test automatic enabling of configuration files of already enabled plugins
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_FILE_ENABLED_INT)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 0)
        for plugin_conf_file_name in YumPluginManager.PLUGINS:
            yum_plugin_config = ConfigParser()
            result = yum_plugin_config.read(
                YumPluginManager.DNF_PLUGIN_DIR + "/" + plugin_conf_file_name + ".conf"
            )
            self.assertGreater(len(result), 0)
            is_plugin_enabled = yum_plugin_config.getint("main", "enabled")
            self.assertEqual(is_plugin_enabled, 1)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_enabled_dnf_plugin_bool(self, mock_get_distribution):
        """
        Test automatic enabling of configuration files of already enabled plugins
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_FILE_ENABLED_BOOL)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 0)
        for plugin_conf_file_name in YumPluginManager.PLUGINS:
            yum_plugin_config = ConfigParser()
            result = yum_plugin_config.read(
                YumPluginManager.DNF_PLUGIN_DIR + "/" + plugin_conf_file_name + ".conf"
            )
            self.assertGreater(len(result), 0)
            # The file was not modified. We have to read value with with getboolean()
            is_plugin_enabled = yum_plugin_config.getboolean("main", "enabled")
            self.assertEqual(is_plugin_enabled, True)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_dnf_plugin_with_invalid_values(self, mock_get_distribution):
        """
        Test automatic enabling of configuration files of already enabled plugins
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_FILE_INVALID_VALUE)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 2)
        for plugin_conf_file_name in YumPluginManager.PLUGINS:
            yum_plugin_config = ConfigParser()
            result = yum_plugin_config.read(
                YumPluginManager.DNF_PLUGIN_DIR + "/" + plugin_conf_file_name + ".conf"
            )
            self.assertGreater(len(result), 0)
            is_plugin_enabled = yum_plugin_config.getint("main", "enabled")
            self.assertEqual(is_plugin_enabled, 1)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_dnf_plugin_with_wrong_conf(self, mock_get_distribution):
        """
        Test automatic enabling of configuration files of already plugins with wrong values in conf file.
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_FILE_WRONG_VALUE)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 2)
        for plugin_conf_file_name in YumPluginManager.PLUGINS:
            yum_plugin_config = ConfigParser()
            result = yum_plugin_config.read(
                YumPluginManager.DNF_PLUGIN_DIR + "/" + plugin_conf_file_name + ".conf"
            )
            self.assertGreater(len(result), 0)
            is_plugin_enabled = yum_plugin_config.getint("main", "enabled")
            self.assertEqual(is_plugin_enabled, 1)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_dnf_plugin_with_corrupted_conf_file(self, mock_get_distribution):
        """
        This test only test YumPluginManager.enable_yum_plugins() can survive reading of corrupted
        dnf plugin configuration file
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_FILE_CORRUPTED)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 0)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_dnf_plugin_with_missing_main_section(self, mock_get_distribution):
        """
        This test only test YumPluginManager.enable_yum_plugins() can survive reading of corrupted
        dnf plugin configuration file (missing main section)
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_MISSING_MAIN_SECTION)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 2)

    @patch.object(repolib, "conf", ConfigFromString(config_string=AUTO_ENABLE_PKG_PLUGINS_ENABLED))
    @patch(
        "rhsmlib.facts.hwprobe.HardwareCollector.get_distribution",
        return_value=("", "", "", "", "rhel", []),
    )
    def test_enabling_dnf_plugin_with_missing_enabled_option(self, mock_get_distribution):
        """
        This test only test YumPluginManager.enable_yum_plugins() can survive reading of corrupted
        dnf plugin configuration file (missing option 'enabled')
        """
        self.init_dnf_plugin_conf_files(conf_string=PKG_PLUGIN_CONF_MISSING_ENABLED_OPTION)
        plugin_list = YumPluginManager.enable_pkg_plugins()
        self.assertEqual(len(plugin_list), 2)



# Copyright (c) 2011 - 2012 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
from iniparse.compat import (
    NoOptionError,
    InterpolationMissingOptionError,
    InterpolationDepthError,
    NoSectionError,
)
import os
from tempfile import NamedTemporaryFile
import unittest

from unittest.mock import patch
from rhsm.config import RhsmConfigParser, RhsmHostConfigParser, in_container

TEST_CONFIG = """
[foo]
bar =
bigger_than_32_bit = 21474836470
bigger_than_64_bit = 123456789009876543211234567890
[server]
hostname = server.example.conf
prefix = /candlepin
port = 8443
insecure = 1
proxy_hostname =
proxy_port =
proxy_user =
proxy_password =

[rhsm]
ca_cert_dir = /etc/rhsm/ca-test/
baseurl = https://content.example.com
repomd_gpg_url = https://content.example.com/gpg
repo_ca_cert = %(ca_cert_dir)sredhat-uep-non-default.pem
productCertDir = /etc/pki/product
entitlementCertDir = /etc/pki/entitlement
consumerCertDir = /etc/pki/consumer
report_package_profile = 1
pluginDir = /usr/lib/rhsm-plugins
some_option = %(repo_ca_cert)stest
manage_repos =

[rhsmcertd]
certCheckInterval = 245
"""

# Only properties we're interested in testing in a host config:
HOST_CONFIG = """
[rhsm]
ca_cert_dir = /etc/rhsm/ca
repo_ca_cert = /etc/rhsm/redhat-uep-non-default.pem
entitlementCertDir = /etc/pki/entitlement
"""

HOST_CONFIG_MACRO = """
[rhsm]
ca_cert_dir = /etc/rhsm/ca/
repo_ca_cert = %(ca_cert_dir)sredhat-uep.pem
entitlementCertDir = /etc/pki/entitlement
"""

HOST_CONFIG_NOREPLACE = """
[rhsm]
ca_cert_dir = /etc/rhsm-other/ca
repo_ca_cert = /etc/pki/ca/redhat-uep.pem
entitlementCertDir = /etc/pki/entitlement-other
"""

HOST_CONFIG_ENTDIR_TRAILING_SLASH = """
[rhsm]
ca_cert_dir = /etc/rhsm-other/ca
repo_ca_cert = /etc/pki/ca/redhat-uep.pem
entitlementCertDir = /etc/pki/entitlement/
"""

OLD_CONFIG = """
[foo]
bar =
bigger_than_32_bit = 21474836470
bigger_than_64_bit = 123456789009876543211234567890
[server]
hostname = server.example.conf
prefix = /candlepin
port = 8443
insecure = 1
ca_cert_dir = /etc/rhsm/ca-old/
proxy_hostname =
proxy_port =
proxy_user =
proxy_password =

[rhsm]
baseurl = https://content.example.com
repomd_gpg_url = https://content.example.com/gpg
repo_ca_cert = %(ca_cert_dir)sredhat-uep.pem
productCertDir = /etc/pki/product
entitlementCertDir = /etc/pki/entitlement
consumerCertDir = /etc/pki/consumer
report_package_profile = 1
pluginDir = /usr/lib/rhsm-plugins

[rhsmcertd]
certCheckInterval = 245
"""

BROKEN_CONFIG = """
[foo]
bar =
bigger_than_32_bit = 21474836470
bigger_than_64_bit = 123456789009876543211234567890
[server]
hostname = server.example.conf
prefix = /candlepin
port = 8443
insecure = 1
ca_cert_dir = /etc/rhsm/ca-broken/
proxy_hostname =
proxy_port =
proxy_user =
proxy_password =

[rhsm]
baseurl = https://content.example.com
repomd_gpg_url = https://content.example.com/gpg
repo_ca_cert = %(s %%(ca_cert_dir)sredhat-uep.pem
productCertDir = /etc/pki/product
entitlementCertDir = /etc/pki/entitlement
consumerCertDir = /etc/pki/consumer
report_package_profile = 1
pluginDir = /usr/lib/rhsm-plugins
some_option = %(repo_ca_cert)s-%(consumerCertDir)s-foo

[rhsmcertd]
certCheckInterval = 245
"""

INTERPOLATION_ERROR_CONFIG = """
[foo]
blip = blip_value
[bar]
interp_key = %(blip)s_and_more
[deeper]
deep = level_one
deeper = %(deep)s_1
deepest = %(deeper)s_2
one_more = %(eleven)s
eleven = %(one_more)s
"""

CA_CERT_DIR_CONFIG = """
[rhsm]
ca_cert_dir = /etc/not-default/
repo_ca_cert = %(ca_cert_dir)snon_default.pem
"""

NO_CA_CERT_DIR_CONFIG = """
[rhsm]
repo_ca_cert = %(ca_cert_dir)snon_default.pem
"""


def write_temp_file(data):
    # create a temp file for use as a config file. This should get cleaned
    # up magically at the end of the run.
    fid = NamedTemporaryFile(mode="w+", suffix=".tmp")
    fid.write(data)
    fid.seek(0)
    return fid


class BaseConfigTests(unittest.TestCase):
    def setUp(self):
        self.fid = write_temp_file(self.cfgfile_data)
        self.cfgParser = RhsmConfigParser(self.fid.name)


class HostConfigTests(unittest.TestCase):
    @patch("os.path.exists")
    def test_normal_case(self, exists_mock):
        # Mock that /etc/pki/entitlement-host exists:
        exists_mock.return_value = True
        temp_file = write_temp_file(HOST_CONFIG)
        config = RhsmHostConfigParser(temp_file.name)
        self.assertEqual("/etc/rhsm-host/ca", config.get("rhsm", "ca_cert_dir"))
        self.assertEqual("/etc/rhsm-host/redhat-uep-non-default.pem", config.get("rhsm", "repo_ca_cert"))
        self.assertEqual("/etc/pki/entitlement-host", config.get("rhsm", "entitlementCertDir"))

    @patch("os.path.exists")
    def test_host_config_regular_entcertdir(self, exists_mock):
        # Mock that /etc/pki/entitlement-host does not exist,
        # our setting should be left alone.
        exists_mock.return_value = False
        temp_file = write_temp_file(HOST_CONFIG)
        config = RhsmHostConfigParser(temp_file.name)
        self.assertEqual("/etc/rhsm-host/ca", config.get("rhsm", "ca_cert_dir"))
        self.assertEqual("/etc/rhsm-host/redhat-uep-non-default.pem", config.get("rhsm", "repo_ca_cert"))
        self.assertEqual("/etc/pki/entitlement", config.get("rhsm", "entitlementCertDir"))

    @patch("os.path.exists")
    def test_repo_ca_cert_macro(self, exists_mock):
        # Mock that /etc/pki/entitlement-host exists:
        exists_mock.return_value = True
        temp_file = write_temp_file(HOST_CONFIG_MACRO)
        config = RhsmHostConfigParser(temp_file.name)
        self.assertEqual("/etc/rhsm-host/ca/", config.get("rhsm", "ca_cert_dir"))
        self.assertEqual("/etc/rhsm-host/ca/redhat-uep.pem", config.get("rhsm", "repo_ca_cert"))

    @patch("os.path.exists")
    def test_no_replacements(self, exists_mock):
        # Mock that /etc/pki/entitlement-host exists:
        exists_mock.return_value = True
        temp_file = write_temp_file(HOST_CONFIG_NOREPLACE)
        config = RhsmHostConfigParser(temp_file.name)
        self.assertEqual("/etc/rhsm-other/ca", config.get("rhsm", "ca_cert_dir"))
        self.assertEqual("/etc/pki/ca/redhat-uep.pem", config.get("rhsm", "repo_ca_cert"))
        self.assertEqual("/etc/pki/entitlement-other", config.get("rhsm", "entitlementCertDir"))

    @patch("os.path.exists")
    def test_ent_dir_trailing_slash(self, exists_mock):
        # Mock that /etc/pki/entitlement-host exists:
        exists_mock.return_value = True
        temp_file = write_temp_file(HOST_CONFIG_ENTDIR_TRAILING_SLASH)
        config = RhsmHostConfigParser(temp_file.name)
        self.assertEqual("/etc/pki/entitlement-host", config.get("rhsm", "entitlementCertDir"))


class ConfigTests(BaseConfigTests):
    cfgfile_data = TEST_CONFIG

    def testRead(self):
        self.assertEqual(self.cfgParser.get("server", "hostname"), "server.example.conf")

    def testSet(self):
        self.cfgParser.set("rhsm", "baseurl", "cod")
        self.assertEqual(self.cfgParser.get("rhsm", "baseurl"), "cod")

    def test_get(self):
        value = self.cfgParser.get("rhsm", "baseurl")
        self.assertEqual("https://content.example.com", value)

    def test_get_empty(self):
        value = self.cfgParser.get("foo", "bar")
        self.assertEqual("", value)

    def test_get_repomd_gpg_url(self):
        value = self.cfgParser.get("rhsm", "repomd_gpg_url")
        self.assertEqual("https://content.example.com/gpg", value)

    def test_get_repo_ca_cert(self):
        value = self.cfgParser.get("rhsm", "repo_ca_cert")
        self.assertEqual("/etc/rhsm/ca-test/redhat-uep-non-default.pem", value)

    def test_has_default_true(self):
        value = self.cfgParser.has_default("server", "hostname")
        self.assertTrue(value)

    def test_has_default_false(self):
        value = self.cfgParser.has_default("foo", "port")
        self.assertFalse(value)

    def test_is_default_true(self):
        value = self.cfgParser.is_default("server", "hostname", "subscription.rhsm.redhat.com")
        self.assertTrue(value)

    def test_is_default_false(self):
        value = self.cfgParser.is_default("server", "hostname", "localhost")
        self.assertFalse(value)

    def test_get_default_camel_case(self):
        value = self.cfgParser.get_default("rhsmcertd", "certCheckInterval")
        self.assertEqual("240", value)

    def test_get_default(self):
        value = self.cfgParser.get_default("rhsmcertd", "certcheckinterval")
        self.assertEqual("240", value)

    def test_get_int(self):
        value = self.cfgParser.get_int("server", "port")
        self.assertTrue(isinstance(value, int))
        self.assertEqual(8443, value)

    def test_interpolation(self):
        value = self.cfgParser.get("rhsm", "repo_ca_cert")
        self.assertEqual("/etc/rhsm/ca-test/redhat-uep-non-default.pem", value)

    def test_get_item_does_not_exist(self):
        self.assertRaises(NoOptionError, self.cfgParser.get, "rhsm", "this_isnt_a_thing")

    def test_get_int_un_set(self):
        value = self.cfgParser.get_int("server", "proxy_port")
        self.assertEqual(None, value)

    def test_get_int_does_not_exist(self):
        self.assertRaises(NoOptionError, self.cfgParser.get_int, "rhsm", "this_isnt_a_thing")

    def test_get_int_not_an_int(self):
        self.assertRaises(ValueError, self.cfgParser.get_int, "rhsm", "baseurl")

    def test_get_int_big_int(self):
        value = self.cfgParser.get_int("foo", "bigger_than_32_bit")
        self.assertEqual(21474836470, value)
        value = self.cfgParser.get_int("foo", "bigger_than_64_bit")
        self.assertEqual(123456789009876543211234567890, value)


class SomeOptionConfigTest(BaseConfigTests):
    cfgfile_data = TEST_CONFIG

    def test_interpolation_nested(self):
        value = self.cfgParser.get("rhsm", "some_option")
        self.assertEqual("/etc/rhsm/ca-test/redhat-uep-non-default.pemtest", value)


class BlankWithDefaultConfigTest(BaseConfigTests):
    cfgfile_data = TEST_CONFIG

    def test_get_blank_config_file_entry(self):
        default_returned = False
        for (name, value) in self.cfgParser.items("rhsm"):
            if name == "manage_repos":
                default_returned = value == "1"
        self.assertTrue(default_returned)


class OldConfigTests(ConfigTests):
    cfgfile_data = OLD_CONFIG

    def test_get_repo_ca_cert(self):
        value = self.cfgParser.get("rhsm", "repo_ca_cert")
        # for old style configs, with ca_cert_dir in 'server', we
        #  expect to ignore that (as we did by accident before), and
        # get the default of '/etc/rhsm/ca'
        self.assertEqual("/etc/rhsm/ca/redhat-uep.pem", value)

    def test_interpolation(self):
        value = self.cfgParser.get("rhsm", "repo_ca_cert")
        self.assertEqual("/etc/rhsm/ca/redhat-uep.pem", value)


# this config has an invalid value for repo_ca_cert
class BrokenConfigTests(ConfigTests):
    cfgfile_data = BROKEN_CONFIG

    # our repo_ca_cert is busted, so expect an exception
    def test_get_repo_ca_cert(self):
        self.assertRaises(InterpolationMissingOptionError, self.cfgParser.get, "rhsm", "repo_ca_cert")

    # interp fails, expect interpolation exception
    def test_interpolation(self):
        self.assertRaises(InterpolationMissingOptionError, self.cfgParser.get, "rhsm", "repo_ca_cert")

    def test_nested_interpolation(self):
        self.assertRaises(InterpolationMissingOptionError, self.cfgParser.get, "rhsm", "some_option")

    def test_not_a_section(self):
        self.assertRaises(NoSectionError, self.cfgParser.get, "not_a_section", "not_an_option")


class InterpErrorTests(BaseConfigTests):
    cfgfile_data = INTERPOLATION_ERROR_CONFIG

    def test_get_interp_key(self):
        self.assertRaises(InterpolationMissingOptionError, self.cfgParser.get, "bar", "interp_key")

    def test_deeper(self):
        value = self.cfgParser.get("deeper", "deeper")
        self.assertEqual("level_one_1", value)

    def test_deepest(self):
        value = self.cfgParser.get("deeper", "deepest")
        self.assertEqual("level_one_1_2", value)

    def test_one_more(self):
        self.assertRaises(InterpolationDepthError, self.cfgParser.get, "deeper", "one_more")


class CaCertDirTests(BaseConfigTests):
    cfgfile_data = CA_CERT_DIR_CONFIG

    def test_get_repo_ca_cert(self):
        value = self.cfgParser.get("rhsm", "repo_ca_cert")
        self.assertEqual("/etc/not-default/non_default.pem", value)


class NoCaCertDirTests(BaseConfigTests):
    cfgfile_data = NO_CA_CERT_DIR_CONFIG

    def test_get_repo_ca_cert(self):
        value = self.cfgParser.get("rhsm", "repo_ca_cert")
        self.assertEqual("/etc/rhsm/ca/non_default.pem", value)


class InContainerTests(unittest.TestCase):
    def test_in_container_off(self):
        # This must be set in the test environment by CI in order to know for
        # sure one way or the other whether we really are in a container.
        # It must NOT be set to something other than the truth of whether we are
        # in a container otherwise this test will no longer be meaningful
        really_in_container = True if os.environ.get("SUBMAN_TEST_IN_CONTAINER", False) else False
        with patch.dict(os.environ, {"SMDEV_CONTAINER_OFF": "True"}):
            self.assertFalse(in_container())
        self.assertEqual(in_container(), really_in_container)

    @patch.dict("os.environ", {}, clear=True)
    @patch("os.path.exists")
    def test_existing_dotcontainer(self, exists_mock):
        def exists_file(path):
            return path == "/run/.containerenv"

        exists_mock.side_effect = exists_file
        self.assertTrue(in_container())

    @patch.dict("os.environ", {}, clear=True)
    @patch("os.path.exists")
    def test_existing_dotdockerenv(self, exists_mock):
        def exists_file(path):
            return path == "/.dockerenv"

        exists_mock.side_effect = exists_file
        self.assertTrue(in_container())

    @patch.dict("os.environ", {}, clear=True)
    @patch("os.path.exists")
    def test_existing_rhsm_host(self, exists_mock):
        def exists_file(path):
            return path == "/etc/rhsm-host/"

        exists_mock.side_effect = exists_file
        self.assertTrue(in_container())

    @patch.dict(
        "os.environ",
        {
            "KUBERNETES_PORT": "tcp://10.0.0.1:443",
            "KUBERNETES_SERVICE_HOST": "10.0.0.1",
            "KUBERNETES_SERVICE_PORT": "443",
            "container": "oci",
        },
        clear=True,
    )
    @patch("os.path.exists")
    def test_ocp(self, exists_mock):
        def exists_file(path):
            return path == "/run/.containerenv"

        exists_mock.side_effect = exists_file
        self.assertFalse(in_container())



# Copyright (c) 2017 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

"""
This module contains implementation of D-Bus object representing system purpose.
It uses interface: com.redhat.RHSM1.Syspurpose and path:
/com/redhat/RHSM1/Syspurpose
"""

import dbus
import json
import logging

from rhsmlib.dbus import constants, base_object, util, dbus_utils
from rhsmlib.services import syspurpose
from syspurpose.files import SyspurposeStore

from subscription_manager.injectioninit import init_dep_injection
from subscription_manager.i18n import ungettext
from subscription_manager.i18n import Locale

init_dep_injection()

log = logging.getLogger(__name__)


class ThreeWayMergeConflict(dbus.DBusException):
    """
    Raise this exception, when client application tries to
    """

    _dbus_error_name = "%s.Error" % constants.SYSPURPOSE_INTERFACE
    include_traceback = False
    severity = "warn"

    def __init__(self, conflict_fields):
        """
        Initialize this exception
        :param conflict_fields: dictionary with conflicted attributes.
            The key is attribute and value is current value set on server.
        """
        self.conflict_fields = conflict_fields

    def __str__(self):
        """
        Text representation of exception
        :return: string of exception
        """
        conflicts = []
        for key, value in self.conflict_fields.items():
            conflicts.append(
                '{conflict_attr} of "{existing_value}"'.format(
                    conflict_attr=key,
                    existing_value=value,
                )
            )
        conflict_msg = ", ".join(conflicts)
        return ungettext(
            "Warning: The following field was recently set "
            "for this system by the entitlement server "
            "administrator: {conflict_msg}",
            "Warning: The following fields were recently set "
            "for this system by the entitlement server "
            "administrator: {conflict_msg}",
            len(conflicts),
        ).format(conflict_msg=conflict_msg)


class SyspurposeDBusObject(base_object.BaseObject):
    """
    A D-Bus object interacting with subscription-manager to get
    information about current system purpose.
    """

    default_dbus_path = constants.SYSPURPOSE_DBUS_PATH
    interface_name = constants.SYSPURPOSE_INTERFACE

    def __init__(self, conn=None, object_path=None, bus_name=None):
        super(SyspurposeDBusObject, self).__init__(conn=conn, object_path=object_path, bus_name=bus_name)

    @util.dbus_service_method(
        constants.SYSPURPOSE_INTERFACE,
        in_signature="s",
        out_signature="s",
    )
    @util.dbus_handle_sender
    @util.dbus_handle_exceptions
    def GetSyspurpose(self, locale, sender=None):
        """
        D-Bus method for getting current system purpose
        :param locale: string with locale
        :param sender:
        :return: json representation of system purpose contents
        """
        syspurpose_path = "/etc/rhsm/syspurpose/syspurpose.json"

        locale = dbus_utils.dbus_to_python(locale, expected_type=str)
        Locale.set(locale)

        syspurpose_store = SyspurposeStore.read(syspurpose_path)

        try:
            contents = syspurpose_store.contents
        except Exception as err:
            raise dbus.DBusException(str(err))

        return json.dumps(contents)

    @util.dbus_service_method(
        constants.SYSPURPOSE_INTERFACE,
        in_signature="s",
        out_signature="s",
    )
    @util.dbus_handle_exceptions
    def GetSyspurposeStatus(self, locale, sender=None):
        """
        D-Bus method for getting system purpose status
        :param locale: string representing locale
        :param sender: object representing application which called this method
        :return:
        """
        locale = dbus_utils.dbus_to_python(locale, expected_type=str)
        Locale.set(locale)
        cp = self.build_uep({})
        system_purpose = syspurpose.Syspurpose(cp)
        syspurpose_status = system_purpose.get_syspurpose_status()["status"]
        return system_purpose.get_overall_status(syspurpose_status)

    @util.dbus_service_method(
        constants.SYSPURPOSE_INTERFACE,
        in_signature="s",
        out_signature="s",
    )
    @util.dbus_handle_exceptions
    def GetValidFields(self, locale, sender=None):
        """
        Method for getting valid syspurpose attributes and values
        :param locale: string with locale
        :param sender: object representing application which called this method
        :return: string representing dictionary with valid fields
        """
        locale = dbus_utils.dbus_to_python(locale, expected_type=str)
        Locale.set(locale)
        cp = self.build_uep({})
        system_purpose = syspurpose.Syspurpose(cp)
        valid_fields = system_purpose.get_owner_syspurpose_valid_fields()
        if valid_fields is None:
            # When it is not possible to get valid fields, then raise exception
            if self.is_registered() is False:
                raise dbus.DBusException(
                    "Unable to get system purpose valid fields. System is not registered.",
                )
            else:
                raise dbus.DBusException(
                    "Unable to get system purpose valid fields.",
                )
        else:
            return json.dumps(valid_fields)

    @util.dbus_service_method(
        constants.SYSPURPOSE_INTERFACE,
        in_signature="a{sv}s",
        out_signature="s",
    )
    @util.dbus_handle_sender
    @util.dbus_handle_exceptions
    def SetSyspurpose(self, syspurpose_values, locale, sender):
        """
        Set syspurpose values
        :param syspurpose_values: Dictionary with all syspurpose values
        :param locale: String with locale
        :param sender: Object representing client application that called this method
        :return: String with successfully set syspurpose values
        """
        syspurpose_values = dbus_utils.dbus_to_python(syspurpose_values, expected_type=dict)
        locale = dbus_utils.dbus_to_python(locale, expected_type=str)
        Locale.set(locale)

        cp = self.build_uep({})
        system_purpose = syspurpose.Syspurpose(cp)
        new_syspurpose_values = system_purpose.set_syspurpose_values(syspurpose_values)

        # Check if there was any conflict during three-way merge
        conflicts = {}
        for key, value in new_syspurpose_values.items():
            if key in syspurpose_values and syspurpose_values[key] != value:
                conflicts[key] = value
        if len(conflicts) > 0:
            raise ThreeWayMergeConflict(conflict_fields=conflicts)

        return json.dumps(new_syspurpose_values)

    @util.dbus_service_signal(
        constants.SYSPURPOSE_INTERFACE,
        signature="",
    )
    @util.dbus_handle_exceptions
    def SyspurposeChanged(self):
        """
        Signal fired, when system purpose is created/deleted/changed
        :param sender:
        :return: None
        """
        log.debug("D-Bus signal %s emitted" % constants.SYSPURPOSE_INTERFACE)
        return None



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import logging
import os

from rhsm import connection
from rhsm.connection import ProxyException

from subscription_manager import syspurposelib
from subscription_manager.cli import system_exit
from subscription_manager.cli_command.cli import CliCommand, ERR_NOT_REGISTERED_CODE, ERR_NOT_REGISTERED_MSG
from subscription_manager.i18n import ungettext, ugettext as _
from subscription_manager.syspurposelib import get_syspurpose_valid_fields
from subscription_manager.utils import friendly_join

from syspurpose.files import SyncedStore, post_process_received_data

log = logging.getLogger(__name__)

SP_CONFLICT_MESSAGE = _(
    'Warning: A {attr} of "{download_value}" was recently set for this system '
    "by the entitlement server administrator.\n{advice}"
)
SP_ADVICE = _("If you'd like to overwrite the server side change please run: {command}")

# TRANSLATORS: this refers to a deprecated command
DEPRECATED_COMMAND_MESSAGE = _("Deprecated, see 'syspurpose'")


class AbstractSyspurposeCommand(CliCommand):
    """
    Abstract command for manipulating an attribute of system purpose.
    """

    def __init__(
        self,
        name,
        subparser,
        shortdesc=None,
        primary=False,
        attr=None,
        commands=("set", "unset", "show", "list"),
    ):
        # set 'subparser' before calling the parent constructor, as it will
        # (indirectly) call _create_argparser(), which our reimplementation uses
        self.subparser = subparser
        if self.subparser is None:
            # this syspurpose command is a deprecated top-level subcommand,
            # so change its description to be a deprecated text
            shortdesc = DEPRECATED_COMMAND_MESSAGE
        super(AbstractSyspurposeCommand, self).__init__(name, shortdesc=shortdesc, primary=primary)
        self.commands = commands
        self.attr = attr

        self.store = None

        if "set" in commands:
            self.parser.add_argument(
                "--set",
                dest="set",
                help=_("set {attr} of system purpose").format(attr=attr),
            )
        if "unset" in commands:
            self.parser.add_argument(
                "--unset",
                dest="unset",
                action="store_true",
                help=_("unset {attr} of system purpose").format(attr=attr),
            )
        if "add" in commands:
            self.parser.add_argument(
                "--add",
                dest="to_add",
                action="append",
                default=[],
                help=_("add an item to the list ({attr}).").format(attr=attr),
            )
        if "remove" in commands:
            self.parser.add_argument(
                "--remove",
                dest="to_remove",
                action="append",
                default=[],
                help=_("remove an item from the list ({attr}).").format(attr=attr),
            )
        if "show" in commands:
            self.parser.add_argument(
                "--show",
                dest="show",
                action="store_true",
                help=_("show this system's current {attr}").format(attr=attr),
            )
        if "list" in commands:
            self.parser.add_argument(
                "--list",
                dest="list",
                action="store_true",
                help=_("list all {attr} available").format(attr=attr),
            )

    def __getattr__(self, name):
        """
        This custom __getattr__() reimplementation is used to lookup attributes
        in the parent syspurpose command, if set; this is done in case the
        current command is a subcommand of 'syspurpose', so
        SyspurposeCommand._do_command() will set a reference to itself as
        'syspurpose_command' attribute.
        """
        syspurpose_command = self.__dict__.get("syspurpose_command", None)
        if syspurpose_command is not None:
            return getattr(syspurpose_command, name)
        raise AttributeError

    def _validate_options(self):
        to_set = getattr(self.options, "set", None)
        to_unset = getattr(self.options, "unset", None)
        to_add = getattr(self.options, "to_add", None)
        to_remove = getattr(self.options, "to_remove", None)
        to_show = getattr(self.options, "show", None)

        if to_set:
            self.options.set = self.options.set.strip()
        if to_add:
            self.options.to_add = [x.strip() for x in self.options.to_add if isinstance(x, str)]
        if to_remove:
            self.options.to_remove = [x.strip() for x in self.options.to_remove if isinstance(x, str)]
        if (to_set or to_add or to_remove) and to_unset:
            system_exit(os.EX_USAGE, _("--unset cannot be used with --set, --add, or --remove"))
        if to_add and to_remove:
            system_exit(os.EX_USAGE, _("--add cannot be used with --remove"))

        if not self.is_registered():
            if self.options.list:
                if self.options.token and not self.options.username:
                    pass
                elif self.options.token and self.options.username:
                    system_exit(os.EX_USAGE, _("Error: you can specify --username or --token not both"))
                elif not self.options.username or not self.options.password:
                    system_exit(
                        os.EX_USAGE,
                        _(
                            "Error: you must register or specify --username and --password to list {attr}"
                        ).format(attr=self.attr),
                    )
            elif to_unset or to_set or to_add or to_remove or to_show:
                pass
            else:
                system_exit(ERR_NOT_REGISTERED_CODE, ERR_NOT_REGISTERED_MSG)

        if self.is_registered() and (
            getattr(self.options, "username", None)
            or getattr(self.options, "password", None)
            or getattr(self.options, "token", None)
            or getattr(self.options, "org", None)
        ):
            system_exit(
                os.EX_USAGE,
                _(
                    "Error: --username, --password, --token and --org "
                    "can be used only on unregistered systems"
                ),
            )

    def _get_valid_fields(self):
        """
        Try to get valid fields from server
        :return: Dictionary with valid fields
        """
        valid_fields = {}
        if self.is_registered():
            # When system is registered, then try to get valid fields from cache file
            try:
                valid_fields = get_syspurpose_valid_fields(uep=self.cp, identity=self.identity)
            except ProxyException as exc:
                system_exit(os.EX_UNAVAILABLE, exc)
        elif self.options.username and self.options.password and self.cp is not None:
            # Try to get current organization key. It is property of OrgCommand.
            # Every Syspurpose command has to be subclass of OrgCommand too
            # must have used credentials in command if not registered to proceed
            try:
                org_key = self.org
                server_response = self.cp.getOwnerSyspurposeValidFields(org_key)
            except connection.RestlibException as rest_err:
                log.warning(
                    "Unable to get list of valid fields using REST API: {rest_err}".format(rest_err=rest_err)
                )
                system_exit(os.EX_SOFTWARE, rest_err)
            except ProxyException as exc:
                system_exit(os.EX_UNAVAILABLE, exc)
            else:
                if "systemPurposeAttributes" in server_response:
                    server_response = post_process_received_data(server_response)
                    valid_fields = server_response["systemPurposeAttributes"]
        return valid_fields

    def _is_provided_value_valid(self, value):
        """
        Try to validate provided value. Check if the value is included in valid fields.
        If the value is not provided in the valid fields and we can connect candlepin server,
        then print some warning.
        :param value: provided value on CLI
        :return: True if the value is valid; otherwise return False
        """
        invalid_values = self._are_provided_values_valid([value])
        return len(invalid_values) == 0

    def _are_provided_values_valid(self, values):
        """
        Try to validate the provided values. Check if all the values are included in valid fields.
        If any of the values is not provided in the valid fields and we can connect candlepin
        server, then print some warning.
        :param values: provided values on CLI
        :return: list of invalid values
        """

        # First check if the the value is in the valid_fields.  Comparison is case insensitive.
        invalid_values = []
        valid_fields = self._get_valid_fields()
        if self.attr in valid_fields:
            for value in values:
                if all([x.casefold() != value.casefold() for x in valid_fields[self.attr]]):
                    invalid_values.append(value)
        invalid_values_len = len(invalid_values)

        # When there are values not in the valid fields, then try to print some warning,
        # when the system is registered or username & password was provided as CLI option.
        # When the system is not registered and no username & password was provided, then
        # these values will be set silently.
        if invalid_values_len > 0:
            if (
                self.is_registered()
                or (self.options.username and self.options.password)
                or self.options.token
            ):
                if len(valid_fields.get(self.attr, [])) > 0:
                    # TRANSLATORS: this is used to quote a string
                    quoted_values = [_('"{value}"').format(value=value) for value in invalid_values]
                    printable_values = friendly_join(quoted_values)
                    print(
                        ungettext(
                            "Warning: Provided value {vals} is not included in the list of valid values",
                            "Warning: Provided values {vals} are not included in the list of valid values",
                            invalid_values_len,
                        ).format(vals=printable_values)
                    )
                    self._print_valid_values(valid_fields)
                else:
                    print(
                        _(
                            "Warning: This organization does not have any subscriptions that provide a "
                            'system purpose "{attr}".  This setting will not influence auto-attaching '
                            "subscriptions."
                        ).format(attr=self.attr)
                    )

        return invalid_values

    def set(self):
        """
        Try to set new syspurpose attribute
        """
        self._set(self.options.set)
        self._is_provided_value_valid(self.options.set)
        success_msg = _('{attr} set to "{val}".').format(attr=self.attr, val=self.options.set)
        self._check_result(
            expectation=lambda res: res.get(self.attr) == self.options.set,
            success_msg=success_msg,
            command='subscription-manager syspurpose {name} --set "{val}"'.format(
                name=self.name, val=self.options.set
            ),
            attr=self.attr,
        )

    def _set(self, to_set):
        if self.store:
            self.store.set(self.attr, to_set)
        else:
            log.debug("Not setting syspurpose attribute {attr} (store not set)".format(attr=self.attr))

    def unset(self):
        self._unset()
        success_msg = _("{attr} unset.").format(attr=self.attr)
        self._check_result(
            expectation=lambda res: res.get(self.attr) in ["", None, []],
            success_msg=success_msg,
            command="subscription-manager syspurpose {name} --unset".format(name=self.name),
            attr=self.attr,
        )

    def _unset(self):
        if self.store:
            self.store.unset(self.attr)
        else:
            log.debug("Not unsetting syspurpose attribute {attr} (store not set)".format(attr=self.attr))

    def add(self):
        self._add(self.options.to_add)
        self._are_provided_values_valid(self.options.to_add)
        success_msg = _("{attr} updated.").format(attr=self.name)
        # When there is several options to add, then format of command is following
        # subscription-manager command --add opt1 --add opt2
        options = ['"' + option + '"' for option in self.options.to_add]
        to_add = "--add " + " --add ".join(options)
        command = "subscription-manager syspurpose {name} ".format(name=self.name) + to_add
        self._check_result(
            expectation=lambda res: all(x in res.get("addons", []) for x in self.options.to_add),
            success_msg=success_msg,
            command=command,
            attr=self.attr,
        )

    def _add(self, to_add):
        if not isinstance(to_add, list):
            to_add = [to_add]

        if self.store:
            for item in to_add:
                self.store.add(self.attr, item)
        else:
            log.debug("Not adding syspurpose attribute {attr} (store not set)".format(attr=self.attr))

    def remove(self):
        self._remove(self.options.to_remove)
        success_msg = _("{attr} updated.").format(attr=self.name.capitalize())
        options = ['"' + option + '"' for option in self.options.to_remove]
        # When there is several options to remove, then format of command is following
        # subscription-manager syspurpose command --remove opt1 --remove opt2
        to_remove = "--remove " + " --remove ".join(options)
        command = "subscription-manager syspurpose {name} ".format(name=self.name) + to_remove
        self._check_result(
            expectation=lambda res: all(x not in res.get("addons", []) for x in self.options.to_remove),
            success_msg=success_msg,
            command=command,
            attr=self.attr,
        )

    def _remove(self, to_remove):
        if not isinstance(to_remove, list):
            to_remove = [to_remove]

        if self.store:
            for item in to_remove:
                self.store.remove(self.attr, item)
        else:
            log.debug("Not removing syspurpose attribute {attr} (store not set)".format(attr=self.attr))

    def show(self):
        if self.is_registered():
            syspurpose = self.sync().result
        else:
            syspurpose = syspurposelib.read_syspurpose()
        if syspurpose is not None and self.attr in syspurpose and syspurpose[self.attr]:
            val = syspurpose[self.attr]
            values = val if not isinstance(val, list) else ", ".join(val)
            print(_("Current {name}: {val}").format(name=self.name.capitalize(), val=values))
        else:
            print(_("{name} not set.").format(name=self.name.capitalize()))

    def _print_valid_values(self, valid_fields):
        """
        Print list of valid values for current syspurpose attribute
        :param valid_fields:
        :return: None
        """
        for valid_value in valid_fields[self.attr]:
            if len(valid_value) > 0:
                print(" - {value}".format(value=valid_value))

    def list(self):
        valid_fields = self._get_valid_fields()
        if self.attr in valid_fields:
            if len(valid_fields[self.attr]) > 0:
                line = "+-------------------------------------------+"
                print(line)
                translated_string = _("Available {syspurpose_attr}").format(syspurpose_attr=self.attr)
                # Print translated string (the length could be different) in the center of the line
                line_len = len(line)
                trans_str_len = len(translated_string)
                empty_space_len = int((line_len - trans_str_len) / 2)
                print(empty_space_len * " " + translated_string)
                print(line)
                # Print values
                self._print_valid_values(valid_fields)
            else:
                print(
                    _(
                        'There are no available values for the system purpose "{syspurpose_attr}" '
                        "from the available subscriptions in this "
                        "organization."
                    ).format(syspurpose_attr=self.attr)
                )
        else:
            print(
                _(
                    "Unable to get the list of valid values for the system purpose " '"{syspurpose_attr}".'
                ).format(syspurpose_attr=self.attr)
            )

    def sync(self):
        return syspurposelib.SyspurposeSyncActionCommand().perform(
            include_result=True, passthrough_gone=True
        )[1]

    def _do_command(self):
        self._validate_options()

        self.cp = None
        try:
            # If we have a username/password, we're going to use that, otherwise
            # we'll use the identity certificate. We already know one or the other
            # exists:
            if self.options.token:
                try:
                    self.cp = self.cp_provider.get_keycloak_auth_cp(self.options.token)
                except Exception as err:
                    log.error(
                        'unable to connect to candlepin server using token: "{token}", err: {err}'.format(
                            token=self.options.token, err=err
                        )
                    )
                    print(_("Unable to connect to server using token"))
            elif self.options.username and self.options.password:
                self.cp_provider.set_user_pass(self.options.username, self.options.password)
                self.cp = self.cp_provider.get_basic_auth_cp()
            else:
                # get an UEP as consumer
                if self.is_registered():
                    self.cp = self.cp_provider.get_consumer_auth_cp()
        except connection.RestlibException as err:
            log.exception(err)
            if getattr(self.options, "list", None):
                log.error(
                    "Error: Unable to retrieve {attr} from server: {err}".format(attr=self.attr, err=err)
                )
                system_exit(os.EX_SOFTWARE, err)
            else:
                log.debug(
                    "Error: Unable to retrieve {attr} from server: {err}".format(attr=self.attr, err=err)
                )
        except Exception as err:
            log.debug("Error: Unable to retrieve {attr} from server: {err}".format(attr=self.attr, err=err))

        self.store = SyncedStore(uep=self.cp, consumer_uuid=self.identity.uuid)

        if getattr(self.options, "unset", None):
            self.unset()
        elif getattr(self.options, "set", None):
            self.set()
        elif hasattr(self.options, "to_add") and len(self.options.to_add) > 0:
            self.add()
        elif hasattr(self.options, "to_remove") and len(self.options.to_remove) > 0:
            self.remove()
        elif getattr(self.options, "list", None):
            self.list()
        elif getattr(self.options, "show", None):
            self.show()
        else:
            self.show()

    def _create_argparser(self):
        if self.subparser is None:
            # Without a subparser, which means it is a standalone command
            return super(AbstractSyspurposeCommand, self)._create_argparser()

        # This string is similar to what _get_usage() returns; we cannot use
        # _get_usage() as it prints the subcommand name as well, and the created
        # ArgumentParser is a subparser (so there is a parent parser already
        # printing the subcommand).
        usage = _("%(prog)s [OPTIONS]")
        return self.subparser.add_parser(
            self.name, description=self.shortdesc, usage=usage, help=self.shortdesc
        )

    def check_syspurpose_support(self, attr):
        if self.is_registered() and not self.cp.has_capability("syspurpose"):
            print(
                _(
                    "Note: The currently configured entitlement server does "
                    "not support System Purpose {attr}.".format(attr=attr)
                )
            )

    def _check_result(self, expectation, success_msg, command, attr):
        if self.store:
            self.store.sync()
            result = self.store.get_cached_contents()
        else:
            result = {}
        if result and not expectation(result):
            advice = SP_ADVICE.format(command=command)
            value = result[attr]
            msg = _(SP_CONFLICT_MESSAGE.format(attr=attr, download_value=value, advice=advice))
            system_exit(os.EX_SOFTWARE, msg)
        else:
            print(success_msg)



from ..test_managercli import TestCliProxyCommand
from subscription_manager import managercli


class TestOwnersCommand(TestCliProxyCommand):
    command_class = managercli.OwnersCommand

    def test_main_server_url(self):
        server_url = "https://subscription.rhsm.redhat.com/subscription"
        self.cc.main(["--serverurl", server_url])

    def test_insecure(self):
        self.cc.main(["--insecure"])

    def test_token_(self):
        self.cc.main(["--token", "eyJhbGciOiJSUzI1NiIsInR5cCIg"])



import os

from ..test_managercli import TestCliProxyCommand
from subscription_manager import managercli
from subscription_manager.cli_command.abstract_syspurpose import AbstractSyspurposeCommand
import subscription_manager.injection as inj

from ..stubs import StubUEP
from ..fixture import Capture, SubManFixture

from unittest.mock import patch, Mock, MagicMock


class TestSyspurposeCommand(TestCliProxyCommand):
    command_class = managercli.RoleCommand

    def setUp(self):
        synced_store_patch = patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
        self.synced_store_mock = synced_store_patch.start()
        self.addCleanup(self.synced_store_mock)
        syspurpose_patch = patch("syspurpose.files.SyncedStore")
        sp_patch = syspurpose_patch.start()
        self.addCleanup(sp_patch.stop)
        super(TestSyspurposeCommand, self).setUp(False)
        self.cc = self.command_class()
        self.cc.cp = StubUEP()
        self.cc.cp.registered_consumer_info["role"] = None
        self.cc.cp._capabilities = ["syspurpose"]

    def test_show_option(self):
        self.cc.main(["--show"])


class TestRoleCommand(TestCliProxyCommand):
    command_class = managercli.RoleCommand

    def setUp(self):
        synced_store_patch = patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
        self.synced_store_mock = synced_store_patch.start()
        self.addCleanup(self.synced_store_mock)
        syspurpose_patch = patch("syspurpose.files.SyncedStore")
        sp_patch = syspurpose_patch.start()
        self.addCleanup(sp_patch.stop)
        super(TestRoleCommand, self).setUp(False)
        self.cc = self.command_class()
        self.cc.cp = StubUEP()
        self.cc.cp.registered_consumer_info["role"] = None
        self.cc.cp._capabilities = ["syspurpose"]

    def test_list_username_password_org(self):
        self.cc.is_registered = Mock(return_value=False)
        self.cc.main(["--username", "admin", "--password", "secret", "--org", "one", "--list"])

    def test_list_username_password(self):
        self.cc.is_registered = Mock(return_value=False)
        self.cc.main(["--username", "admin", "--password", "secret", "--list"])

    def test_list_only_username(self):
        self.cc.options = Mock()
        self.cc.is_registered = Mock(return_value=False)
        self.cc.options.set = False
        self.cc.options.unset = False
        self.cc.options.to_add = False
        self.cc.options.to_remove = False
        self.cc.options.list = True
        self.cc.options.token = None
        self.cc.options.username = "admin"
        self.cc.options.password = None
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_list_username_and_token(self):
        self.cc.options = Mock()
        self.cc.is_registered = Mock(return_value=False)
        self.cc.options.set = False
        self.cc.options.unset = False
        self.cc.options.to_add = False
        self.cc.options.to_remove = False
        self.cc.options.list = True
        self.cc.options.token = "TOKEN"
        self.cc.options.username = "admin"
        self.cc.options.password = "secret"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_wrong_options_syspurpose_role(self):
        """It is possible to use --set or --unset options. It's not possible to use both of them together."""
        self.cc.options = Mock()
        self.cc.options.set = "Foo"
        self.cc.options.unset = True
        self.cc.options.to_add = False
        self.cc.options.to_remove = False
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_username_on_registered_system(self):
        """Argument --username cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.username = "admin"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_password_on_registered_system(self):
        """Argument --password cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.password = "secret"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_token_on_registered_system(self):
        """Argument --token cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.token = "TOKEN"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_org_on_registered_system(self):
        """Argument --org cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.org = "organization"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    def test_main_no_args(self, mock_syspurpose):
        # It is necessary to mock SyspurposeStore for test function of parent class
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.local_contents = {"role": "Foo"}

        with patch.object(AbstractSyspurposeCommand, "check_syspurpose_support", Mock(return_value=None)):
            super(TestRoleCommand, self).test_main_no_args()

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    def test_main_empty_args(self, mock_syspurpose):
        # It is necessary to mock SyspurposeStore for test function of parent class
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.local_contents = {"role": "Foo"}

        with patch.object(AbstractSyspurposeCommand, "check_syspurpose_support", Mock(return_value=None)):
            super(TestRoleCommand, self).test_main_empty_args()

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    @patch("subscription_manager.syspurposelib.SyspurposeSyncActionCommand")
    def test_display_valid_syspurpose_role(self, mock_syspurpose_sync, mock_syspurpose):
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.contents = {"role": "Foo"}

        self.cc.options = Mock(spec=["set", "unset"])
        self.cc.options.set = None
        self.cc.options.unset = False

        mock_syspurpose_sync.return_value = Mock()
        instance_mock_syspurpose_sync = mock_syspurpose_sync.return_value

        mock_sync_result = Mock()
        mock_sync_result.result = {"role": "Foo"}
        instance_mock_syspurpose_sync.perform = Mock(return_value=({}, mock_sync_result))

        with Capture() as cap:
            self.cc._do_command()
        self.assertIn("Current Role: Foo", cap.out)

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    def test_display_none_syspurpose_role(self, mock_syspurpose):
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.contents = {"role": None}

        self.cc.options = Mock(spec=["set", "unset"])
        self.cc.options.set = None
        self.cc.options.unset = False
        with Capture() as cap:
            self.cc._do_command()
        self.assertIn("Role not set", cap.out)

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    def test_display_nonexisting_syspurpose_role(self, mock_syspurpose):
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.contents = {}

        self.cc.options = Mock(spec=["set", "unset"])
        self.cc.options.set = None
        self.cc.options.unset = False

        with Capture() as cap:
            self.cc._do_command()
        self.assertIn("Role not set.", cap.out)

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    @patch("subscription_manager.syspurposelib.SyspurposeSyncActionCommand")
    def test_setting_syspurpose_role(self, mock_syspurpose_sync, mock_syspurpose):
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.contents = {}
        instance_syspurpose_store.set = MagicMock(return_value=True)
        instance_syspurpose_store.write = MagicMock(return_value=None)
        instance_syspurpose_store.get_cached_contents = Mock(return_value={"role": "Foo"})
        mock_syspurpose.return_value = instance_syspurpose_store

        mock_syspurpose_sync.return_value = Mock()
        instance_mock_syspurpose_sync = mock_syspurpose_sync.return_value
        instance_mock_syspurpose_sync.perform = Mock(return_value=({}, {"role": "Foo"}))

        self.cc.options = Mock(spec=["set", "unset"])
        self.cc.options.set = "Foo"
        self.cc.options.unset = False

        with Capture() as cap:
            self.cc._do_command()

        self.assertIn('role set to "Foo"', cap.out)
        instance_syspurpose_store.set.assert_called_once_with("role", "Foo")
        instance_syspurpose_store.sync.assert_called_once()

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    @patch("subscription_manager.syspurposelib.SyspurposeSyncActionCommand")
    def test_setting_conflicting_syspurpose_role(self, mock_syspurpose_sync, mock_syspurpose):
        """
        Test the case, when there is conflict with value set by administrator on server
        """
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.contents = {}
        instance_syspurpose_store.set = MagicMock(return_value=True)
        instance_syspurpose_store.write = MagicMock(return_value=None)
        instance_syspurpose_store.get_cached_contents = Mock(return_value={"role": "Foo"})
        mock_syspurpose.return_value = instance_syspurpose_store

        mock_syspurpose_sync.return_value = Mock()
        instance_mock_syspurpose_sync = mock_syspurpose_sync.return_value
        instance_mock_syspurpose_sync.perform = Mock(return_value=({}, {"role": "Foo"}))

        self.cc.options = Mock(spec=["set", "unset"])
        self.cc.options.set = "Bar"
        self.cc.options.unset = False

        with Capture() as cap:
            with self.assertRaises(SystemExit) as cm:
                self.cc._do_command()
            self.assertEqual(cm.exception.code, 70)

        self.assertIn(
            'Warning: A role of "Foo" was recently set for this system '
            "by the entitlement server administrator.",
            cap.err,
        )
        self.assertIn(
            "If you'd like to overwrite the server side change please run: "
            'subscription-manager syspurpose role --set "Bar"',
            cap.err,
        )

    @patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
    @patch("subscription_manager.syspurposelib.SyspurposeSyncActionCommand")
    def test_unsetting_syspurpose_role(self, mock_syspurpose_sync, mock_syspurpose):
        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        instance_syspurpose_store = mock_syspurpose.read.return_value
        instance_syspurpose_store.contents = {"role": "Foo"}
        instance_syspurpose_store.unset = MagicMock(return_value=True)
        instance_syspurpose_store.write = MagicMock(return_value=None)
        instance_syspurpose_store.get_cached_contents = MagicMock(return_value={})
        mock_syspurpose.return_value = instance_syspurpose_store

        mock_syspurpose_sync.return_value = Mock()
        instance_mock_syspurpose_sync = mock_syspurpose_sync.return_value
        instance_mock_syspurpose_sync.perform = Mock(return_value=({}, {"role": ""}))

        self.cc.options = Mock(spec=["set", "unset"])
        self.cc.options.set = None
        self.cc.options.unset = True
        with Capture() as cap:
            self.cc._do_command()

        self.assertIn("role unset", cap.out)
        instance_syspurpose_store.unset.assert_called_once_with("role")
        instance_syspurpose_store.sync.assert_called_once()

    def test_is_provided_value_valid(self):
        self.cc = AbstractSyspurposeCommand("role", None, shortdesc="role", primary=False, attr="role")
        self.cc._get_valid_fields = Mock()
        self.cc._get_valid_fields.return_value = {"role": ["Welcome to the Machine"]}
        res = self.cc._is_provided_value_valid("wElcOme To The mAChiNE")
        self.assertTrue(res)


class InvalidOrg(Exception):
    """
    Small Exception to be used in TestRoleOrgCheckingCommand; it will be
    passed through handle_exception().
    """

    def __init__(self, org_key):
        self.org_key = org_key


class OrgCheckingStubUEP(StubUEP):
    """
    An improved StubUEP that returns a specific list of valid organizations,
    and raises InvalidOrg in getOwnerSyspurposeValidFields() in case the
    requested organization is not one of the allowed ones. This will allow us
    to check which AbstractSyspurposeCommand._get_valid_fields() is actually
    passing to UEP.

    The extra implementation on top of StubUEP is the minimal one to check
    AbstractSyspurposeCommand in TestRoleOrgCheckingCommand.
    """

    def __init__(self, allowed_orgs):
        super(OrgCheckingStubUEP, self).__init__()
        self.allowed_orgs = allowed_orgs

    def getOwnerList(self, username):
        return [{"key": org} for org in self.allowed_orgs]

    def getOwnerSyspurposeValidFields(self, org_key):
        if org_key not in self.allowed_orgs:
            raise InvalidOrg(org_key)
        return super(OrgCheckingStubUEP, self).getOwnerSyspurposeValidFields(org_key)


class TestRoleOrgCheckingCommand(SubManFixture):
    common_args = ["--list", "--username", "test", "--password", "test"]

    def do_reraise(self, *args, **kwargs):
        # handle_exception() was called with (msg, ex); we need to reraise
        # 'ex', effectively bypassing handle_exception()
        if len(args) >= 2:
            raise args[1] from None

    def set_orgs_for_stub_cp_provider(self, orgs):
        cp_provider = inj.require(inj.CP_PROVIDER)
        cp_provider.consumer_auth_cp = OrgCheckingStubUEP(orgs)
        cp_provider.basic_auth_cp = OrgCheckingStubUEP(orgs)
        cp_provider.no_auth_cp = OrgCheckingStubUEP(orgs)

    def setUp(self):
        synced_store_patch = patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
        self.synced_store_mock = synced_store_patch.start()
        self.addCleanup(self.synced_store_mock)
        syspurpose_patch = patch("syspurpose.files.SyncedStore")
        sp_patch = syspurpose_patch.start()
        self.addCleanup(sp_patch.stop)
        handle_exception_patch = patch("subscription_manager.cli_command.cli.handle_exception")
        self.handle_exception_mock = handle_exception_patch.start()
        self.handle_exception_mock.side_effect = self.do_reraise
        self.addCleanup(self.handle_exception_mock)
        super(TestRoleOrgCheckingCommand, self).setUp()
        self.cc = managercli.RoleCommand()
        self.cc.is_registered = Mock(return_value=False)

    def test_no_org(self):
        self.set_orgs_for_stub_cp_provider([])
        with self.assertRaises(InvalidOrg):
            self.cc.main(self.common_args + ["--org", "foo"])

    def test_single_org_same(self):
        self.set_orgs_for_stub_cp_provider(["org1"])
        self.cc.main(self.common_args + ["--org", "org1"])

    def test_single_org_different(self):
        self.set_orgs_for_stub_cp_provider(["org1"])
        with self.assertRaises(InvalidOrg):
            self.cc.main(self.common_args + ["--org", "foo"])

    def test_multiple_orgs_same(self):
        self.set_orgs_for_stub_cp_provider(["org1", "org2"])
        self.cc.main(self.common_args + ["--org", "org1"])

    def test_multiple_orgs_different(self):
        self.set_orgs_for_stub_cp_provider(["org1", "org2"])
        with self.assertRaises(InvalidOrg):
            self.cc.main(self.common_args + ["--org", "foo"])



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import logging
import os
import readline
import signal

import rhsm.connection as connection
import subscription_manager.injection as inj

from argparse import SUPPRESS

from rhsm.connection import RemoteServerException
from rhsm.https import ssl
from rhsm.utils import LiveStatusMessage

from rhsmlib.facts.hwprobe import ClassicCheck
from rhsmlib.services import attach, unregister, register, exceptions

from subscription_manager import identity
from subscription_manager.branding import get_branding
from subscription_manager.cli import system_exit
from subscription_manager.cli_command.cli import handle_exception, conf
from subscription_manager.cli_command.environments import MULTI_ENV
from subscription_manager.cli_command.list import show_autosubscribe_output
from subscription_manager.cli_command.user_pass import UserPassCommand
from subscription_manager.entcertlib import CONTENT_ACCESS_CERT_CAPABILITY
from subscription_manager.i18n import ugettext as _
from subscription_manager.utils import (
    restart_virt_who,
    print_error,
    get_supported_resources,
    is_simple_content_access,
    is_interactive,
    is_process_running,
)
from subscription_manager.cli_command.environments import check_set_environment_names
from subscription_manager.exceptions import ExceptionMapper

log = logging.getLogger(__name__)


class RegisterCommand(UserPassCommand):
    def __init__(self):
        shortdesc = get_branding().CLI_REGISTER

        super(RegisterCommand, self).__init__("register", shortdesc, True)

        self._add_url_options()
        self.parser.add_argument(
            "--baseurl",
            dest="base_url",
            default=None,
            help=_("base URL for content in form of https://hostname:port/prefix"),
        )
        self.parser.add_argument(
            "--type",
            dest="consumertype",
            default="system",
            metavar="UNITTYPE",
            help=SUPPRESS,
        )
        self.parser.add_argument(
            "--name",
            dest="consumername",
            metavar="SYSTEMNAME",
            help=_("name of the system to register, defaults to the hostname"),
        )
        self.parser.add_argument(
            "--consumerid",
            dest="consumerid",
            metavar="SYSTEMID",
            help=_("the existing system data is pulled from the server"),
        )
        self.parser.add_argument(
            "--org",
            dest="org",
            metavar="ORG_KEY",
            help=_("register with one of multiple organizations for the user, using organization key"),
        )
        self.parser.add_argument(
            "--environments",
            dest="environments",
            help=_(
                "register with a specific environment (single value) or multiple environments "
                "(a comma-separated list) in the destination org. The ability to use multiple "
                "environments is controlled by the entitlement server"
            ),
        )
        self.parser.add_argument(
            "--release",
            dest="release",
            help=_("set a release version"),
        )
        self.parser.add_argument(
            "--autosubscribe",
            action="store_true",
            help=_("Deprecated, see --auto-attach"),
        )
        self.parser.add_argument(
            "--auto-attach",
            action="store_true",
            dest="autoattach",
            help=_("automatically attach compatible subscriptions to this system"),
        )
        self.parser.add_argument(
            "--force",
            action="store_true",
            help=_("include an implicit attempt to unregister before registering a new system identity"),
        )
        self.parser.add_argument(
            "--activationkey",
            action="append",
            dest="activation_keys",
            help=_("activation key to use for registration (can be specified more than once)"),
        )
        self.parser.add_argument(
            "--servicelevel",
            dest="service_level",
            help=_("system preference used when subscribing automatically, requires --auto-attach"),
        )

    def _validate_options(self):
        self.autoattach = self.options.autosubscribe or self.options.autoattach
        if self.is_registered() and not self.options.force:
            system_exit(os.EX_USAGE, _("This system is already registered. Use --force to override"))
        elif self.options.consumername == "":
            system_exit(os.EX_USAGE, _("Error: system name can not be empty."))
        elif (self.options.username or self.options.token) and self.options.activation_keys:
            system_exit(os.EX_USAGE, _("Error: Activation keys do not require user credentials."))
        elif self.options.consumerid and self.options.activation_keys:
            system_exit(
                os.EX_USAGE, _("Error: Activation keys can not be used with previously registered IDs.")
            )
        elif self.options.environments and self.options.activation_keys:
            system_exit(os.EX_USAGE, _("Error: Activation keys do not allow environments to be specified."))
        elif self.autoattach and self.options.activation_keys:
            system_exit(os.EX_USAGE, _("Error: Activation keys cannot be used with --auto-attach."))
        # 746259: Don't allow the user to pass in an empty string as an activation key
        elif self.options.activation_keys and "" in self.options.activation_keys:
            system_exit(os.EX_USAGE, _("Error: Must specify an activation key"))
        elif self.options.service_level and not self.autoattach:
            system_exit(os.EX_USAGE, _("Error: Must use --auto-attach with --servicelevel."))
        elif self.options.activation_keys and not self.options.org:
            system_exit(os.EX_USAGE, _("Error: Must provide --org with activation keys."))
        elif self.options.force and self.options.consumerid:
            system_exit(
                os.EX_USAGE,
                _(
                    "Error: Can not force registration while attempting to recover registration "
                    "with consumerid. Please use --force without --consumerid to re-register or "
                    "use the clean command and try again without --force."
                ),
            )
        # 1485008: allow registration, when --type=RHUI (many of KBase articles describe using RHUI not rhui)
        elif self.options.consumertype and not (
            self.options.consumertype.lower() == "rhui" or self.options.consumertype == "system"
        ):
            system_exit(os.EX_USAGE, _("Error: The --type option has been deprecated and may not be used."))
        if self.options.environments:
            if not self.cp.has_capability(MULTI_ENV) and "," in self.options.environments:
                system_exit(os.EX_USAGE, _("The entitlement server does not allow multiple environments"))

    def persist_server_options(self):
        """
        If the user provides a --serverurl or --baseurl, we want to persist it
        to the config file so that future commands will use the value.
        """
        return True

    def _do_auto_attach(self, consumer):
        """
        Try to do auto-attach, when it was requested using --auto-attach CLI option
        :return: None
        """

        # Do not try to do auto-attach, when simple content access mode is used
        # Only print info message to stdout
        if is_simple_content_access(uep=self.cp, identity=self.identity):
            self._print_ignore_auto_attach_message()
            return

        if "serviceLevel" not in consumer and self.options.service_level:
            system_exit(
                os.EX_UNAVAILABLE,
                _(
                    "Error: The --servicelevel option is not supported "
                    "by the server. Did not complete your request."
                ),
            )
        try:
            # We don't call auto_attach with self.option.service_level, because it has been already
            # set during service.register() call
            attach.AttachService(self.cp).attach_auto(service_level=None)
        except connection.RestlibException as rest_lib_err:
            mapped_message: str = ExceptionMapper().get_message(rest_lib_err)
            print_error(mapped_message)
        except Exception:
            log.exception("Auto-attach failed")
            raise

    def _upload_profile_blocking(self, consumer: dict) -> None:
        """
        Try to upload DNF profile to server
        """
        with LiveStatusMessage(_("Uploading DNF profile")):
            try:
                profile_mgr = inj.require(inj.PROFILE_MANAGER)
                # 767265: always force an upload of the packages when registering
                profile_mgr.update_check(self.cp, consumer["uuid"], force=True)
            except RemoteServerException as err:
                # When it is not possible to upload profile ATM, then print only error about this
                # to rhsm.log. The rhsmcertd will try to upload it next time.
                log.error("Unable to upload profile: {err!s}".format(err=err))

    def _upload_profile(self, consumer: dict) -> None:
        """
        Try to upload DNF profile to server, when it is supported by server. This method
        tries to "outsource" this activity to rhsmcertd first. When it is not possible due to
        various reasons, then we try to do it ourselves in blocking way.
        """
        # First try to get PID of rhsmcertd from lock file
        try:
            with open("/var/lock/subsys/rhsmcertd", "r") as lock_file:
                rhsmcertd_pid = int(lock_file.readline())
        except (IOError, ValueError) as err:
            log.info(f"Unable to read rhsmcertd lock file: {err}")
        else:
            if is_process_running("rhsmcertd", rhsmcertd_pid) is True:
                # This will only send SIGUSR1 signal, which triggers gathering and uploading
                # of DNF profile by rhsmcertd. We try to "outsource" this activity to rhsmcertd
                # server to not block registration process.
                # Note: rhsmcertd tries to upload profile using Python script and this script
                # is always triggered with --force-upload CLI option. We ignore report_package_config
                # configure option here due to BZ: 767265
                log.debug("Sending SIGUSR1 signal to rhsmcertd process")
                try:
                    os.kill(rhsmcertd_pid, signal.SIGUSR1)
                except ProcessLookupError as err:
                    # When rhsmcertd process was terminated between calling is_process_running()
                    # and sending signal using kill(), then fallback to uploading profile from
                    # current process
                    log.debug(f"Unable to send signal SIGUSR1 to rhsmcertd process {rhsmcertd_pid}: {err}")
                    self._upload_profile_blocking(consumer)
            else:
                # When rhsmcertd process is not running, then fallback to uploading profile from
                # current process
                log.info(f"rhsmcertd process with given PID: {rhsmcertd_pid} is not running")
                self._upload_profile_blocking(consumer)

    def _do_command(self):
        """
        Executes the command.
        """

        self.log_client_version()

        # Always warn the user if registered to old RHN/Spacewalk
        if ClassicCheck().is_registered_with_classic():
            print(get_branding().REGISTERED_TO_OTHER_WARNING)

        self._validate_options()

        # gather installed products info
        self.installed_mgr = inj.require(inj.INSTALLED_PRODUCTS_MANAGER)

        previously_registered = False
        if self.is_registered() and self.options.force:
            previously_registered = True
            # First let's try to un-register previous consumer; if this fails
            # we'll let the error bubble up, so that we don't blindly re-register.
            # managerlib.unregister handles the special case that the consumer has already been removed.
            old_uuid = self.identity.uuid

            print(
                _("Unregistering from: {hostname}:{port}{prefix}").format(
                    hostname=conf["server"]["hostname"],
                    port=conf["server"]["port"],
                    prefix=conf["server"]["prefix"],
                )
            )
            try:
                unregister.UnregisterService(self.cp).unregister()
                self.entitlement_dir.__init__()
                self.product_dir.__init__()
                log.info("--force specified, unregistered old consumer: {old_uuid}".format(old_uuid=old_uuid))
                print(_("The system with UUID {old_uuid} has been unregistered").format(old_uuid=old_uuid))
            except ssl.SSLError as e:
                # since the user can override serverurl for register, a common use case
                # is to try to switch servers using register --force... However, this
                # normally cannot successfully unregister since the servers are different.
                handle_exception("Unregister failed: {e}".format(e=e), e)
            except Exception as e:
                handle_exception("Unregister failed", e)

        self.cp_provider.clean()
        if previously_registered:
            print(_("All local data removed"))

        # Proceed with new registration:
        try:
            if self.options.token:
                admin_cp = self.cp_provider.get_keycloak_auth_cp(self.options.token)
            elif not self.options.activation_keys:
                hostname = conf["server"]["hostname"]
                if ":" in hostname:
                    normalized_hostname = "[{hostname}]".format(hostname=hostname)
                else:
                    normalized_hostname = hostname
                print(
                    _("Registering to: {hostname}:{port}{prefix}").format(
                        hostname=normalized_hostname,
                        port=conf["server"]["port"],
                        prefix=conf["server"]["prefix"],
                    )
                )
                self.cp_provider.set_user_pass(self.username, self.password)
                admin_cp = self.cp_provider.get_basic_auth_cp()
            else:
                admin_cp = self.cp_provider.get_no_auth_cp()

            # This is blocking and not async, which aside from blocking here, also
            # means things like following name owner changes gets weird.
            service = register.RegisterService(admin_cp)

            if self.options.consumerid:
                log.debug("Registering as existing consumer: {id}".format(id=self.options.consumerid))
                consumer = service.register(None, consumerid=self.options.consumerid)
            else:
                if self.options.org:
                    owner_key = self.options.org
                else:
                    owner_key = service.determine_owner_key(
                        username=self.username, get_owner_cb=self._get_owner_cb, no_owner_cb=self._no_owner_cb
                    )
                environment_ids = self._process_environments(admin_cp, owner_key)
                consumer = service.register(
                    owner_key,
                    activation_keys=self.options.activation_keys,
                    environments=environment_ids,
                    force=self.options.force,
                    name=self.options.consumername,
                    consumer_type=self.options.consumertype,
                    service_level=self.options.service_level,
                )
        except (connection.RestlibException, exceptions.ServiceError) as re:
            log.exception(re)

            system_exit(os.EX_SOFTWARE, re)
        except Exception as e:
            handle_exception(_("Error during registration: {e}").format(e=e), e)
        else:
            consumer_info = identity.ConsumerIdentity(consumer["idCert"]["key"], consumer["idCert"]["cert"])
            print(_("The system has been registered with ID: {id}").format(id=consumer_info.getConsumerId()))
            print(_("The registered system name is: {name}").format(name=consumer_info.getConsumerName()))
            if self.options.service_level:
                print(_("Service level set to: {level}").format(level=self.options.service_level))

        # We have new credentials, restart virt-who
        restart_virt_who()

        # get a new UEP as the consumer
        self.cp = self.cp_provider.get_consumer_auth_cp()

        # log the version of the server we registered to
        self.log_server_version()

        facts = inj.require(inj.FACTS)

        # FIXME: can these cases be replaced with invoking
        # FactsLib (or a FactsManager?)
        # Must update facts to clear out the old ones:
        if self.options.consumerid:
            log.debug("Updating facts")
            #
            # FIXME: Need a ConsumerFacts.sync or update or something
            # TODO: We register, with facts, then update facts again...?
            #       Are we trying to sync potential new or dynamic facts?
            facts.update_check(self.cp, consumer["uuid"], force=True)

        # Facts and installed products went out with the registration request,
        # manually write caches to disk:
        # facts service job now(soon)
        facts.write_cache()
        self.installed_mgr.update_check(self.cp, consumer["uuid"])

        if self.options.release:
            # TODO: grab the list of valid options, and check
            self.cp.updateConsumer(consumer["uuid"], release=self.options.release)

        if self.autoattach:
            self._do_auto_attach(consumer)

        if (
            self.options.consumerid
            or self.options.activation_keys
            or self.autoattach
            or self.cp.has_capability(CONTENT_ACCESS_CERT_CAPABILITY)
        ):
            log.debug("System registered, updating entitlements if needed")
            # update certs, repos, and caches.
            # FIXME: aside from the overhead, should this be cert_action_client.update?
            self.entcertlib.update()

        self._upload_profile(consumer)

        subscribed = 0
        if self.options.activation_keys or self.autoattach:
            # update with the latest cert info
            self.sorter = inj.require(inj.CERT_SORTER)
            self.sorter.force_cert_check()
            subscribed = show_autosubscribe_output(self.cp, self.identity)

        self._request_validity_check()
        return subscribed

    def _prompt_for_environment(self):
        """
        By breaking this code out, we can write cleaner tests
        """
        if not is_interactive():
            system_exit(
                os.EX_USAGE,
                _("Error: --environments is a required parameter in non-interactive mode."),
            )

        if self.cp.has_capability(MULTI_ENV):
            environment = input(_("Environments: ")).replace(" ", "")
        else:
            environment = input(_("Environment: ")).strip()
        readline.clear_history()
        return environment or self._prompt_for_environment()

    def _process_environments(self, admin_cp, owner_key):
        """
        Confirms that environment(s) have been chosen if they are supported
        and a choice needs to be made
        """
        supported_resources = get_supported_resources()
        supports_environments = "environments" in supported_resources

        if not supports_environments:
            if self.options.environments is not None:
                system_exit(os.EX_UNAVAILABLE, _("Error: Server does not support environments."))
            return None

        # We have an activation key, so don't need to fill/check the bits
        # related to environments, as they are part of the activation key
        if self.options.activation_keys:
            return None

        all_env_list = admin_cp.getEnvironmentList(owner_key)
        if self.options.environments:
            environments = self.options.environments
        else:
            # If there aren't any environments, don't prompt for one
            if not all_env_list:
                return None

            # If the envronment list is len 1, pick that environment
            if len(all_env_list) == 1:
                log.debug(
                    'Using the only available environment: "{name}"'.format(name=all_env_list[0]["name"])
                )
                return all_env_list[0]["id"]

            env_name_list = [env["name"] for env in all_env_list]
            print(
                _('Hint: Organization "{key}" contains following environments: {list}').format(
                    key=owner_key, list=", ".join(env_name_list)
                )
            )
            environments = self._prompt_for_environment()
            if not self.cp.has_capability(MULTI_ENV) and "," in environments:
                system_exit(os.EX_USAGE, _("The entitlement server does not allow multiple environments"))

        return check_set_environment_names(all_env_list, environments)

    @staticmethod
    def _no_owner_cb(username):
        """
        Method called, when there it no owner in the list of owners for given user
        :return: None
        """
        system_exit(1, _("{name} cannot register with any organizations.").format(name=username))

    def _get_owner_cb(self, owners):
        """
        Callback method used, when it is necessary to specify owner (organization)
        during registration
        :param owners: list of owners (organizations)
        :return:
        """
        # Print list of owners to the console
        org_keys = [owner["key"] for owner in owners]
        print(
            _('Hint: User "{name}" is member of following organizations: {orgs}').format(
                name=self.username, orgs=", ".join(org_keys)
            )
        )

        # Read the owner key from stdin or raise a system error if in a non-interactive session.
        if not is_interactive():
            system_exit(
                os.EX_USAGE,
                _("Error: --org is a required parameter in non-interactive mode."),
            )

        owner_key = None
        while not owner_key:
            owner_key = input(_("Organization: "))
            readline.clear_history()
        return owner_key
